#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <mpi.h>
#include <gsl/gsl_math.h>

#include "allvars.h"
#include "proto.h"

#ifdef SFR_METALS
#include "c_metals.h"
#endif

/*! \file density.c 
 *  \brief SPH density computation and smoothing length determination
 *
 *  This file contains the "first SPH loop", where the SPH densities and some
 *  auxiliary quantities are computed.  There is also functionality that
 *  corrects the smoothing length if needed.
 */


/*! This function computes the local density for each active SPH particle, the
 * number of neighbours in the current smoothing radius, and the divergence
 * and rotation of the velocity field.  The pressure is updated as well.  If a
 * particle with its smoothing region is fully inside the local domain, it is
 * not exported to the other processors. The function also detects particles
 * that have a number of neighbours outside the allowed tolerance range. For
 * these particles, the smoothing length is adjusted accordingly, and the
 * density() computation is called again.  Note that the smoothing length is
 * not allowed to fall below the lower bound set by MinGasHsml (this may mean
 * that one has to deal with substantially more than normal number of
 * neighbours.)
 */
void density(void)
{
  long long ntot, ntotleft;
  int ndone;
  int *noffset, *nbuffer, *nsend, *nsend_local, *numlist, *ndonelist;
  int i, j, n, NNN;
  int npleft;
  int maxfill, source, iter = 0;
  int level, ngrp, sendTask, recvTask;
  int place, nexport;
  double dt_entr, dmax1, dmax2;
  double tstart, tend, tstart_ngb = 0, tend_ngb = 0;
  double sumt, sumcomm;
  double timecomp = 0, timeimbalance = 0, timecommsumm = 0, sumimbalance;
  double timengb, sumtimengb;
  double desnumngb;
  MPI_Status status;

#ifdef REDUCEVISC
  double cs_h, f, hubble_a = 1.0;

  if(All.ComovingIntegrationOn)
    {
      /* Factors for comoving integration of hydro */
      hubble_a = All.Omega0 / (All.Time * All.Time * All.Time)
	+ (1 - All.Omega0 - All.OmegaLambda) / (All.Time * All.Time)
#ifdef DARKENERGY
	+ DarkEnergy_a(All.Time);
#else
	+ All.OmegaLambda;
#endif
    }
#endif

  noffset = malloc(sizeof(int) * NTask);	/* offsets of bunches in common list */
  nbuffer = malloc(sizeof(int) * NTask);
  nsend_local = malloc(sizeof(int) * NTask);
  nsend = malloc(sizeof(int) * NTask * NTask);
  ndonelist = malloc(sizeof(int) * NTask);

#if defined(SFR_METALS) || defined(BLACK_HOLES)
  NNN = NumPart;
#else
  NNN = N_gas;
#endif

  for(n = 0, NumSphUpdate = 0; n < NNN; n++)
    {
#ifdef BLACK_HOLES
      if(P[n].Type == 0 || P[n].Type == 5)
#else
#ifdef SFR
#ifdef SFR_METALS
      if(P[n].Type == 0 || P[n].Type == 6 || P[n].Type == 7)
#else
      if(P[n].Type == 0)
#endif
#endif
#endif
	{
	  PPP[n].Left = PPP[n].Right = 0;

	  if(P[n].Ti_endstep == All.Ti_Current)
	    NumSphUpdate++;

#ifdef BLACK_HOLES
	  P[n].SwallowID = 0;
#endif
#if defined(BLACK_HOLES) && defined(FLTROUNDOFFREDUCTION)
	  if(P[n].Type == 0)
	    SphP[n].dInjected_BH_Energy = SphP[n].Injected_BH_Energy;
#endif
	}
    }

  numlist = malloc(NTask * sizeof(int) * NTask);
  MPI_Allgather(&NumSphUpdate, 1, MPI_INT, numlist, 1, MPI_INT, MPI_COMM_WORLD);
  for(i = 0, ntot = 0; i < NTask; i++)
    ntot += numlist[i];
  free(numlist);



  /* we will repeat the whole thing for those particles where we didn't find enough neighbours */
  do
    {
      i = 0;			/* beginn with this index */
      ntotleft = ntot;		/* particles left for all tasks together */

      while(ntotleft > 0)
	{
	  for(j = 0; j < NTask; j++)
	    nsend_local[j] = 0;

	  /* do local particles and prepare export list */
	  tstart = second();
	  for(nexport = 0, ndone = 0; i < NNN && nexport < All.BunchSizeDensity - NTask; i++)
#ifdef BLACK_HOLES
	    if(P[i].Type == 0 || P[i].Type == 5)
#else
#ifdef SFR
#ifdef SFR_METALS
	    if(P[i].Type == 0 || P[i].Type == 6 || P[i].Type == 7)
#else
	    if(P[i].Type == 0)
#endif
#endif
#endif
	      if(P[i].Ti_endstep == All.Ti_Current)
		{
		  ndone++;

		  for(j = 0; j < NTask; j++)
		    Exportflag[j] = 0;

		  density_evaluate(i, 0);

		  for(j = 0; j < NTask; j++)
		    {
		      if(Exportflag[j])
			{
			  DensDataIn[nexport].Pos[0] = P[i].Pos[0];
			  DensDataIn[nexport].Pos[1] = P[i].Pos[1];
			  DensDataIn[nexport].Pos[2] = P[i].Pos[2];

#if defined(SFR_METALS) || defined(BLACK_HOLES)
			  if(P[i].Type == 0)
#endif
			    {
#ifdef SFR_DECOUPLING
			      DensDataIn[nexport].DensityOld = SphP[i].DensityOld;
			      DensDataIn[nexport].Entropy = SphP[i].Entropy;
#endif
#if defined(BLACK_HOLES)
	      if(P[place].Type != 0)
		{
		  DensDataIn[nexport].Vel[0] = 0;
		  DensDataIn[nexport].Vel[1] = 0;
		  DensDataIn[nexport].Vel[2] = 0;
		}
	      else
#endif
	        {
		DensDataIn[nexport].Vel[0] = SphP[i].VelPred[0];
		DensDataIn[nexport].Vel[1] = SphP[i].VelPred[1];
	        DensDataIn[nexport].Vel[2] = SphP[i].VelPred[2];
	        }

#ifdef WINDS
		DensDataIn[nexport].DelayTime = SphP[i].DelayTime;
#endif
			    }
#ifdef SFR_DECOUPLING
			  else
			    {
			      DensDataIn[nexport].DensityOld = 0;
			    }
#endif

			  DensDataIn[nexport].Hsml = PPP[i].Hsml;
			  DensDataIn[nexport].Index = i;
			  DensDataIn[nexport].Task = j;
			  nexport++;
			  nsend_local[j]++;
			}
		    }
		}
	  tend = second();
	  timecomp += timediff(tstart, tend);

	  qsort(DensDataIn, nexport, sizeof(struct densdata_in), dens_compare_key);

	  for(j = 1, noffset[0] = 0; j < NTask; j++)
	    noffset[j] = noffset[j - 1] + nsend_local[j - 1];

	  tstart = second();

	  MPI_Allgather(nsend_local, NTask, MPI_INT, nsend, NTask, MPI_INT, MPI_COMM_WORLD);

	  tend = second();
	  timeimbalance += timediff(tstart, tend);


	  /* now do the particles that need to be exported */

	  for(level = 1; level < (1 << PTask); level++)
	    {
	      tstart = second();
	      for(j = 0; j < NTask; j++)
		nbuffer[j] = 0;
	      for(ngrp = level; ngrp < (1 << PTask); ngrp++)
		{
		  maxfill = 0;
		  for(j = 0; j < NTask; j++)
		    {
		      if((j ^ ngrp) < NTask)
			if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
			  maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		    }
		  if(maxfill >= All.BunchSizeDensity)
		    break;

		  sendTask = ThisTask;
		  recvTask = ThisTask ^ ngrp;

		  if(recvTask < NTask)
		    {
		      if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
			{
			  /* get the particles */
			  MPI_Sendrecv(&DensDataIn[noffset[recvTask]],
				       nsend_local[recvTask] * sizeof(struct densdata_in), MPI_BYTE,
				       recvTask, TAG_ANY,
				       &DensDataGet[nbuffer[ThisTask]],
				       nsend[recvTask * NTask + ThisTask] * sizeof(struct densdata_in),
				       MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
			}
		    }

		  for(j = 0; j < NTask; j++)
		    if((j ^ ngrp) < NTask)
		      nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
		}
	      tend = second();
	      timecommsumm += timediff(tstart, tend);


	      tstart = second();
	      for(j = 0; j < nbuffer[ThisTask]; j++)
		{
		  density_evaluate(j, 1);
		}
	      tend = second();
	      timecomp += timediff(tstart, tend);

	      /* do a block to measure imbalance */
	      tstart = second();
	      MPI_Barrier(MPI_COMM_WORLD);
	      tend = second();
	      timeimbalance += timediff(tstart, tend);

	      /* get the result */
	      tstart = second();
	      for(j = 0; j < NTask; j++)
		nbuffer[j] = 0;
	      for(ngrp = level; ngrp < (1 << PTask); ngrp++)
		{
		  maxfill = 0;
		  for(j = 0; j < NTask; j++)
		    {
		      if((j ^ ngrp) < NTask)
			if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
			  maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		    }
		  if(maxfill >= All.BunchSizeDensity)
		    break;

		  sendTask = ThisTask;
		  recvTask = ThisTask ^ ngrp;

		  if(recvTask < NTask)
		    {
		      if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
			{
			  /* send the results */
			  MPI_Sendrecv(&DensDataResult[nbuffer[ThisTask]],
				       nsend[recvTask * NTask + ThisTask] * sizeof(struct densdata_out),
				       MPI_BYTE, recvTask, TAG_ANY,
				       &DensDataPartialResult[noffset[recvTask]],
				       nsend_local[recvTask] * sizeof(struct densdata_out),
				       MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);

			  /* add the result to the particles */
			  for(j = 0; j < nsend_local[recvTask]; j++)
			    {
			      source = j + noffset[recvTask];
			      place = DensDataIn[source].Index;

			      PPP[place].NumNgb += DensDataPartialResult[source].Ngb;

#if defined(SFR_METALS) || defined(BLACK_HOLES)
			      if(P[place].Type == 0)
#endif
				{
				  SphP[place].Density += DensDataPartialResult[source].Rho;
				  SphP[place].DivVel += DensDataPartialResult[source].Div;

#ifndef NOGRADHSML
				  SphP[place].DhsmlDensityFactor +=
				    DensDataPartialResult[source].DhsmlDensity;
#endif
				  SphP[place].Rot[0] += DensDataPartialResult[source].Rot[0];
				  SphP[place].Rot[1] += DensDataPartialResult[source].Rot[1];
				  SphP[place].Rot[2] += DensDataPartialResult[source].Rot[2];
#ifdef CONDUCTION
				  SphP[place].SmoothedEntr += DensDataPartialResult[source].SmoothedEntr;
#ifdef CONDUCTION_SATURATION
				  SphP[place].GradEntr[0] += DensDataPartialResult[source].GradEntr[0];
				  SphP[place].GradEntr[1] += DensDataPartialResult[source].GradEntr[1];
				  SphP[place].GradEntr[2] += DensDataPartialResult[source].GradEntr[2];
#endif
#endif

#ifdef SFR_PROMOTION
				  SphP[place].TcoolAvg += DensDataPartialResult[source].TcoolSum;
				  SphP[place].TcoolCount += DensDataPartialResult[source].TcoolCount;
#endif
				}
#ifdef BLACK_HOLES
	      if(P[place].Type == 5)
		{
		  P[place].BH_Density += DensDataPartialResult[source].Rho;
		  P[place].BH_Entropy += DensDataPartialResult[source].SmoothedEntr;
		  P[place].BH_SurroundingGasVel[0] += DensDataPartialResult[source].GasVel[0];
		  P[place].BH_SurroundingGasVel[1] += DensDataPartialResult[source].GasVel[1];
		  P[place].BH_SurroundingGasVel[2] += DensDataPartialResult[source].GasVel[2];
		}
#endif

			    }
			}
		    }

		  for(j = 0; j < NTask; j++)
		    if((j ^ ngrp) < NTask)
		      nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
		}
	      tend = second();
	      timecommsumm += timediff(tstart, tend);

	      level = ngrp - 1;
	    }

	  MPI_Allgather(&ndone, 1, MPI_INT, ndonelist, 1, MPI_INT, MPI_COMM_WORLD);
	  for(j = 0; j < NTask; j++)
	    ntotleft -= ndonelist[j];
	}



      /* do final operations on results */
      tstart = second();
      for(i = 0, npleft = 0; i < NNN; i++)
	{
#ifdef BLACK_HOLES
	  if(P[i].Type == 0 || P[i].Type == 5)
#else
#ifdef SFR
#ifdef SFR_METALS
	  if(P[i].Type == 0 || P[i].Type == 6 || P[i].Type == 7)
#else
	  if(P[i].Type == 0)
#endif
#endif
#endif
	    if(P[i].Ti_endstep == All.Ti_Current)
	      {
#if defined(SFR_METALS) || defined(BLACK_HOLES)
		if(P[i].Type == 0)
#endif
		  {
#ifndef NOGRADHSML
		    SphP[i].DhsmlDensityFactor =
		      1 / (1 + PPP[i].Hsml * SphP[i].DhsmlDensityFactor / (3 * SphP[i].Density));
#endif
		    SphP[i].CurlVel = sqrt(SphP[i].Rot[0] * SphP[i].Rot[0] +
					   SphP[i].Rot[1] * SphP[i].Rot[1] +
					   SphP[i].Rot[2] * SphP[i].Rot[2]) / SphP[i].Density;

		    SphP[i].DivVel /= SphP[i].Density;

#ifdef CONDUCTION
		    SphP[i].SmoothedEntr /= SphP[i].Density;
#ifdef CONDUCTION_SATURATION
		    SphP[i].GradEntr[0] /= SphP[i].Density;
		    SphP[i].GradEntr[1] /= SphP[i].Density;
		    SphP[i].GradEntr[2] /= SphP[i].Density;
#endif
#endif

		    dt_entr =
		      (All.Ti_Current - (P[i].Ti_begstep + P[i].Ti_endstep) / 2) * All.Timebase_interval;

		    SphP[i].Pressure =
		      (SphP[i].Entropy + SphP[i].DtEntropy * dt_entr) * pow(SphP[i].Density, GAMMA);

#ifdef REDUCEVISC
		    if(SphP[i].Density > 0.0)
		      {
			cs_h = sqrt(GAMMA * SphP[i].Pressure / SphP[i].Density) / SphP[i].Hsml;
			f = fabs(SphP[i].DivVel) / (fabs(SphP[i].DivVel) + SphP[i].CurlVel + 0.0001 * cs_h);
			if(All.ComovingIntegrationOn)
			  f *= sqrt(All.Time);
			SphP[i].Dtalpha = -(SphP[i].alpha - All.AlphaMin) * All.DecayTime * cs_h
			  + f * All.ViscSource * DMAX(0.0, -SphP[i].DivVel);
			if(All.ComovingIntegrationOn)
			  SphP[i].Dtalpha /= (hubble_a * All.Time);
		      }
		    else
		      SphP[i].Dtalpha = 0.0;
#endif

#ifdef SFR_PROMOTION
		    if(SphP[i].TcoolCount > 0)
		      SphP[i].TcoolAvg /= SphP[i].TcoolCount;
		    else
		      SphP[i].TcoolAvg = 0;
#endif

		  }

#ifdef BLACK_HOLES
		if(P[i].Type == 5)
		  {
		    if(P[i].BH_Density > 0)
		      {
			P[i].BH_Entropy /= P[i].BH_Density;
			P[i].BH_SurroundingGasVel[0] /= P[i].BH_Density;
			P[i].BH_SurroundingGasVel[1] /= P[i].BH_Density;
			P[i].BH_SurroundingGasVel[2] /= P[i].BH_Density;
		      }
		  }
#endif


		/* now check whether we had enough neighbours */

#ifdef BLACK_HOLES
		if(P[i].Type == 5)
		  desnumngb = All.DesNumNgb * All.BlackHoleNgbFactor;
		else
		  desnumngb = All.DesNumNgb;
#endif

		if(PPP[i].NumNgb < (All.DesNumNgb - All.MaxNumNgbDeviation) ||
		   (PPP[i].NumNgb > (All.DesNumNgb + All.MaxNumNgbDeviation)
		    && PPP[i].Hsml > (1.01 * All.MinGasHsml)))
		  {
		    /* need to redo this particle */
		    npleft++;

		    if(PPP[i].Left > 0 && PPP[i].Right > 0)
		      if((PPP[i].Right - PPP[i].Left) < 1.0e-3 * PPP[i].Left)
			{
			  /* this one should be ok */
			  npleft--;
			  P[i].Ti_endstep = -P[i].Ti_endstep - 1;	/* Mark as inactive */
			  continue;
			}

		    if(PPP[i].NumNgb < (All.DesNumNgb - All.MaxNumNgbDeviation))
		      PPP[i].Left = DMAX(PPP[i].Hsml, PPP[i].Left);
		    else
		      {
			if(PPP[i].Right != 0)
			  {
			    if(PPP[i].Hsml < PPP[i].Right)
			      PPP[i].Right = PPP[i].Hsml;
			  }
			else
			  PPP[i].Right = PPP[i].Hsml;
		      }

		    if(iter >= MAXITER - 10)
		      {
			printf
			  ("i=%d task=%d ID=%d Hsml=%g Left=%g Right=%g Ngbs=%g Right-Left=%g\n   pos=(%g|%g|%g)\n",
			   i, ThisTask, (int) P[i].ID, PPP[i].Hsml, PPP[i].Left, PPP[i].Right,
			   (float) PPP[i].NumNgb, PPP[i].Right - PPP[i].Left, P[i].Pos[0], P[i].Pos[1],
			   P[i].Pos[2]);
			fflush(stdout);
		      }

		    if(PPP[i].Right > 0 && PPP[i].Left > 0)
		      PPP[i].Hsml = pow(0.5 * (pow(PPP[i].Left, 3) + pow(PPP[i].Right, 3)), 1.0 / 3);
		    else
		      {
			if(PPP[i].Right == 0 && PPP[i].Left == 0)
			  endrun(8188);	/* can't occur */

			if(PPP[i].Right == 0 && PPP[i].Left > 0)
			  {
#ifndef NOGRADHSML
			    if(P[i].Type == 0 && fabs(PPP[i].NumNgb - All.DesNumNgb) < 0.5 * All.DesNumNgb)
			      {
				PPP[i].Hsml *=
				  1 - (PPP[i].NumNgb -
				       All.DesNumNgb) / (3 * PPP[i].NumNgb) * SphP[i].DhsmlDensityFactor;
			      }
			    else
			      PPP[i].Hsml *= 1.26;
#else
			    PPP[i].Hsml *= 1.26;
#endif
			  }

			if(PPP[i].Right > 0 && PPP[i].Left == 0)
			  {
#ifndef NOGRADHSML
			    if(P[i].Type == 0 && fabs(PPP[i].NumNgb - All.DesNumNgb) < 0.5 * All.DesNumNgb)
			      {
				PPP[i].Hsml *=
				  1 - (PPP[i].NumNgb -
				       All.DesNumNgb) / (3 * PPP[i].NumNgb) * SphP[i].DhsmlDensityFactor;
			      }
			    else
			      PPP[i].Hsml /= 1.26;
#else
			    PPP[i].Hsml /= 1.26;
#endif
			  }
		      }

		    if(PPP[i].Hsml < All.MinGasHsml)
		      PPP[i].Hsml = All.MinGasHsml;
		  }
		else
		  P[i].Ti_endstep = -P[i].Ti_endstep - 1;	/* Mark as inactive */
	      }
	}
      tend = second();
      timecomp += timediff(tstart, tend);


      numlist = malloc(NTask * sizeof(int) * NTask);
      MPI_Allgather(&npleft, 1, MPI_INT, numlist, 1, MPI_INT, MPI_COMM_WORLD);
      for(i = 0, ntot = 0; i < NTask; i++)
	ntot += numlist[i];
      free(numlist);

      if(ntot > 0)
	{
	  if(iter == 0)
	    tstart_ngb = second();

	  iter++;

	  if(iter > 0 && ThisTask == 0)
	    {
	      printf("ngb iteration %d: need to repeat for %d%09d particles.\n", iter,
		     (int) (ntot / 1000000000), (int) (ntot % 1000000000));
	      fflush(stdout);
	    }

	  if(iter > MAXITER)
	    {
	      printf("failed to converge in neighbour iteration in density()\n");
	      fflush(stdout);
	      endrun(1155);
	    }
	}
      else
	tend_ngb = second();
    }
  while(ntot > 0);


  for(i = 0; i < NNN; i++)
#ifdef BLACK_HOLES
    if(P[i].Type == 0 || P[i].Type == 5)
#else
#ifdef SFR_METALS
    if(P[i].Type == 0 || P[i].Type == 6 || P[i].Type == 7)
#else
    if(P[i].Type == 0)
#endif
#endif
      {
	/* mark as active again */
	if(P[i].Ti_endstep < 0)
	  P[i].Ti_endstep = -P[i].Ti_endstep - 1;
      }

  free(ndonelist);
  free(nsend);
  free(nsend_local);
  free(nbuffer);
  free(noffset);


  /* collect some timing information */

  if(iter > 0)
    timengb = timediff(tstart_ngb, tend_ngb);
  else
    timengb = 0;

  MPI_Reduce(&timengb, &sumtimengb, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&timecomp, &sumt, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&timecommsumm, &sumcomm, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&timeimbalance, &sumimbalance, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

  if(ThisTask == 0)
    {
      All.CPU_HydCompWalk += sumt / NTask;
      All.CPU_HydCommSumm += sumcomm / NTask;
      All.CPU_HydImbalance += sumimbalance / NTask;
      All.CPU_EnsureNgb += sumtimengb / NTask;
    }
}



/*! This function represents the core of the SPH density computation. The
 *  target particle may either be local, or reside in the communication
 *  buffer.
 */
void density_evaluate(int target, int mode)
{
  int j, n;
  int startnode, numngb, numngb_inbox;
  double h, h2, fac, hinv, hinv3, hinv4;
  double rho, divv, wk, dwk;
  double dx, dy, dz, r, r2, u, mass_j;
  double dvx, dvy, dvz;
  double rotv[3];
  FLOAT *pos, *vel;

#if defined(SFR_METALS) || defined(BLACK_HOLES)
  static FLOAT veldummy[3] = { 0, 0, 0 };
#endif

#if defined(CONDUCTION) || defined(BLACK_HOLES)
  double smoothentr;

  smoothentr = 0;
#endif

#ifdef CONDUCTION_SATURATION
  double gradentr[3];
#endif

#ifdef WINDS
  double delaytime;
#endif
#ifndef NOFIXEDMASSINKERNEL
  double weighted_numngb;
#endif
#ifndef NOGRADHSML
  double dhsmlrho;
#endif
#ifdef SFR_DECOUPLING
  FLOAT densityold, entropy;
#endif
#ifdef SFR_PROMOTION
  double tcool_sum;
  int tcool_count;
#endif

#ifdef BLACK_HOLES
 FLOAT gasvel[3];
#endif


#ifdef PERIODIC
  double boxSize, boxHalf;

  boxSize = All.BoxSize;
  boxHalf = 0.5 * All.BoxSize;
#endif


  rho = divv = rotv[0] = rotv[1] = rotv[2] = 0;

#ifndef NOFIXEDMASSINKERNEL
  weighted_numngb = 0;
#endif
#ifndef NOGRADHSML
  dhsmlrho = 0;
#endif

#ifdef CONDUCTION
  smoothentr = 0;
#ifdef CONDUCTION_SATURATION
  gradentr[0] = gradentr[1] = gradentr[2] = 0;
#endif
#endif

#ifdef SFR_PROMOTION
  tcool_sum = 0;
  tcool_count = 0;
#endif

#ifdef BLACK_HOLES
  gasvel[0] = gasvel[1] = gasvel[2] = 0;
#endif

  if(mode == 0)
    {
      pos = P[target].Pos;
#if defined(SFR_METALS) || defined(BLACK_HOLES)
      vel = veldummy;
      if(P[target].Type == 0)
#endif
	{
	  vel = SphP[target].VelPred;
#ifdef WINDS
	  delaytime = SphP[target].DelayTime;
#endif
	}
      h = PPP[target].Hsml;
#ifdef SFR_DECOUPLING
      if(P[target].Type == 0)
	{
	  densityold = SphP[target].DensityOld;
	  entropy = SphP[target].Entropy;
	}
      else
	densityold = 0;
#endif
    }
  else
    {
      pos = DensDataGet[target].Pos;
      vel = DensDataGet[target].Vel;
      h = DensDataGet[target].Hsml;
#ifdef WINDS
      delaytime = DensDataGet[target].DelayTime;
#endif
#ifdef SFR_DECOUPLING
      densityold = DensDataGet[target].DensityOld;
      entropy = DensDataGet[target].Entropy;
#endif
    }


  h2 = h * h;
  hinv = 1.0 / h;
  hinv3 = hinv * hinv * hinv;
  hinv4 = hinv3 * hinv;

  startnode = All.MaxPart;
  numngb = 0;
  do
    {
#ifdef SFR_DECOUPLING
      numngb_inbox = ngb_treefind_variable(&pos[0], h, &startnode, densityold, entropy, &vel[0]);
#else
      numngb_inbox = ngb_treefind_variable(&pos[0], h, &startnode);
#endif

      for(n = 0; n < numngb_inbox; n++)
	{
	  j = Ngblist[n];
#ifdef WINDS
	  if(SphP[j].DelayTime > 0)	/* partner is a wind particle */
	    if(!(delaytime > 0))	/* if I'm not wind, then ignore the wind particle */
	      continue;
#endif
#ifdef BLACK_HOLES
	      if(P[j].Mass == 0)
		continue;
#endif
	  dx = pos[0] - P[j].Pos[0];
	  dy = pos[1] - P[j].Pos[1];
	  dz = pos[2] - P[j].Pos[2];

#ifdef PERIODIC			/*  now find the closest image in the given box size  */
#ifdef LONGBOX
	  if(dx > boxHalf * LONGBOX)
	    dx -= boxSize * LONGBOX;
	  if(dx < -boxHalf * LONGBOX)
	    dx += boxSize * LONGBOX;
#else
	  if(dx > boxHalf)
	    dx -= boxSize;
	  if(dx < -boxHalf)
	    dx += boxSize;
#endif
	  if(dy > boxHalf)
	    dy -= boxSize;
	  if(dy < -boxHalf)
	    dy += boxSize;
	  if(dz > boxHalf)
	    dz -= boxSize;
	  if(dz < -boxHalf)
	    dz += boxSize;
#endif
	  r2 = dx * dx + dy * dy + dz * dz;

	  if(r2 < h2)
	    {
	      numngb++;

	      r = sqrt(r2);

	      u = r * hinv;

	      if(u < 0.5)
		{
		  wk = hinv3 * (2.546479089470 + 15.278874536822 * (u - 1) * u * u);
		  dwk = hinv4 * u * (45.836623610466 * u - 30.557749073644);
		}
	      else
		{
		  wk = hinv3 * 5.092958178941 * (1.0 - u) * (1.0 - u) * (1.0 - u);
		  dwk = hinv4 * (-15.278874536822) * (1.0 - u) * (1.0 - u);
		}

	      mass_j = P[j].Mass;

	      rho += mass_j * wk;

#ifdef BLACK_HOLES
		  gasvel[0] += (mass_j * wk * SphP[j].VelPred[0]);
		  gasvel[1] += (mass_j * wk * SphP[j].VelPred[1]);
		  gasvel[2] += (mass_j * wk * SphP[j].VelPred[2]);
#endif

#if defined(CONDUCTION) || defined(BLACK_HOLES)
		  smoothentr += (mass_j * wk * SphP[j].Entropy);
#ifdef CONDUCTION_SATURATION
	      if(r > 0)
		{
		  gradentr[0] += mass_j * dwk * dx / r * SphP[j].Entropy;
		  gradentr[1] += mass_j * dwk * dy / r * SphP[j].Entropy;
		  gradentr[2] += mass_j * dwk * dz / r * SphP[j].Entropy;
		}
#endif
#endif

#ifndef NOFIXEDMASSINKERNEL
	      weighted_numngb += 4.188790204786 * wk / hinv3;	/* 4.0/3 * PI = 4.188790204786 */
#endif

#ifndef NOGRADHSML
	      dhsmlrho += -mass_j * (3 * hinv * wk + u * dwk);
#endif
	      if(r > 0)
		{
		  fac = mass_j * dwk / r;

		  dvx = vel[0] - SphP[j].VelPred[0];
		  dvy = vel[1] - SphP[j].VelPred[1];
		  dvz = vel[2] - SphP[j].VelPred[2];

		  divv -= fac * (dx * dvx + dy * dvy + dz * dvz);

		  rotv[0] += fac * (dz * dvy - dy * dvz);
		  rotv[1] += fac * (dx * dvz - dz * dvx);
		  rotv[2] += fac * (dy * dvx - dx * dvy);

#ifdef SFR_PROMOTION
		  if(SphP[j].Tcool > 0)
		    {
		      tcool_sum += SphP[j].Tcool;
		      tcool_count++;
		    }
#endif
		}
	    }
	}
    }
  while(startnode >= 0);


  if(mode == 0)
    {
#ifndef NOFIXEDMASSINKERNEL
      PPP[target].NumNgb = weighted_numngb;
#else
      PPP[target].NumNgb = numngb;
#endif

#if defined(SFR_METALS) || defined(BLACK_HOLES)
      if(P[target].Type == 0)
#endif
	{
	  SphP[target].Density = rho;
	  SphP[target].DivVel = divv;
#ifdef CONDUCTION
	  SphP[target].SmoothedEntr = smoothentr;
#ifdef CONDUCTION_SATURATION
	  SphP[target].GradEntr[0] = gradentr[0];
	  SphP[target].GradEntr[1] = gradentr[1];
	  SphP[target].GradEntr[2] = gradentr[2];
#endif
#endif


#ifndef NOGRADHSML
	  SphP[target].DhsmlDensityFactor = dhsmlrho;
#endif
	  SphP[target].Rot[0] = rotv[0];
	  SphP[target].Rot[1] = rotv[1];
	  SphP[target].Rot[2] = rotv[2];

#ifdef SFR_PROMOTION
	  SphP[target].TcoolAvg = tcool_sum;
	  SphP[target].TcoolCount = tcool_count;
#endif
	}
#ifdef BLACK_HOLES
      P[target].BH_Density = rho;
      P[target].BH_Entropy = smoothentr;
      P[target].BH_SurroundingGasVel[0] = gasvel[0];
      P[target].BH_SurroundingGasVel[1] = gasvel[1];
      P[target].BH_SurroundingGasVel[2] = gasvel[2];

/*      printf("Rho %g, Smoothentr %g \n", P[target].BH_Density,  P[target].BH_Entropy); 
	printf("Gas_vel1 %g,  Gas_vel2 %g, Gas_vel3 %g \n", P[target].BH_SurroundingGasVel[0], P[target].BH_SurroundingGasVel[1], P[target].BH_SurroundingGasVel[2]);  */


#endif
    }
  else
    {
      DensDataResult[target].Rho = rho;
      DensDataResult[target].Div = divv;

#ifdef CONDUCTION
      DensDataResult[target].SmoothedEntr = smoothentr;
#ifdef CONDUCTION_SATURATION
      DensDataResult[target].GradEntr[0] = gradentr[0];
      DensDataResult[target].GradEntr[1] = gradentr[1];
      DensDataResult[target].GradEntr[2] = gradentr[2];
#endif
#endif

#ifndef NOFIXEDMASSINKERNEL
      DensDataResult[target].Ngb = weighted_numngb;
#else
      DensDataResult[target].Ngb = numngb;
#endif

#ifndef NOGRADHSML
      DensDataResult[target].DhsmlDensity = dhsmlrho;
#endif
      DensDataResult[target].Rot[0] = rotv[0];
      DensDataResult[target].Rot[1] = rotv[1];
      DensDataResult[target].Rot[2] = rotv[2];

#ifdef SFR_PROMOTION
      DensDataResult[target].TcoolSum = tcool_sum;
      DensDataResult[target].TcoolCount = tcool_count;
#endif

#if defined(CONDUCTION) || defined(BLACK_HOLES)
      DensDataResult[target].SmoothedEntr = smoothentr;
#ifdef CONDUCTION_SATURATION
      DensDataResult[target].GradEntr[0] = gradentr[0];
      DensDataResult[target].GradEntr[1] = gradentr[1];
      DensDataResult[target].GradEntr[2] = gradentr[2];
#endif
#endif

#ifdef BLACK_HOLES
      DensDataResult[target].GasVel[0] = gasvel[0];
      DensDataResult[target].GasVel[1] = gasvel[1];
      DensDataResult[target].GasVel[2] = gasvel[2];
    
/*      printf("Gas_vel1 %g,  Gas_vel2 %g, Gas_vel3 %g \n", DensDataResult[target].GasVel[0], DensDataResult[target].GasVel[1], DensDataResult[target].GasVel[2]);  */

#endif			
    }
}





/*! This routine is a comparison kernel used in a sort routine to group
 *  particles that are exported to the same processor.
 */
int dens_compare_key(const void *a, const void *b)
{
  if(((struct densdata_in *) a)->Task < (((struct densdata_in *) b)->Task))
    return -1;

  if(((struct densdata_in *) a)->Task > (((struct densdata_in *) b)->Task))
    return +1;

  return 0;
}
