#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include <gsl/gsl_math.h>

#include "allvars.h"
#include "proto.h"

/*! \file fof.c
 *  \brief parallel FoF group finder
 */

#ifdef FOF

static struct export_fof
{
  FLOAT Pos[3];
  int Type;
  int NearestIndex;
  int NearestTask;
  long long GrIndex;
}
 *ExportBuf, *ImportBuf;

static struct id_list
{
  long long GrIndex;
  int ID;
  int Type;
  int Task;
  int Index;
  FLOAT Density;
}
 *local_ids, *export_ids;


static double LinkL;
static int Ngroups, TotNgroups, Nids;
static long long TotNids;
static int Nghosts, Nexport, Nlocal;
static int Grid;

static int *local_toGo, *toGo, *export_offset, *import_offset;
static int *Len, *Head, *Next, *Tail, *GridNext;
static int *GroupLen, *GroupOffset, *GroupLenType;
static int *GridFirst;
static long long *GrIndex;
static int *GroupIDs;
static char *GridFlag;

static int *fof_nearest_index;
static int *fof_nearest_task;
static float *fof_nearest_distance;
static float *fof_nearest_hsml;



void fof_fof(int num)
{
  int i, ndm, ndmtot, links, largestgroup;

  if(ThisTask == 0)
    {
      printf("\nBegin to compute FoF group catalogues...\n");
      fflush(stdout);
    }


  for(i = 0, ndm = 0; i < NumPart; i++)
    if(P[i].Type == 1)
      ndm++;

  MPI_Allreduce(&ndm, &ndmtot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  LinkL = LINKLENGTH * All.BoxSize / pow(ndmtot, 1.0 / 3);

  if(ThisTask == 0)
    {
      printf("\nComoving linking length: %g \n\n", LinkL);
      fflush(stdout);
    }


  fof_nearest_index = malloc(sizeof(int) * NumPart);
  fof_nearest_task = malloc(sizeof(int) * NumPart);


  fof_find_nearest_dmparticle();

  local_toGo = malloc(sizeof(int) * NTask);
  toGo = malloc(sizeof(int) * NTask * NTask);
  export_offset = malloc(sizeof(int) * NTask);
  import_offset = malloc(sizeof(int) * NTask);

  fof_import_ghosts();

  Head = malloc((NumPart + Nghosts) * sizeof(int));
  GrIndex = malloc((NumPart + Nghosts) * sizeof(long long));

  Len = malloc((NumPart + Nghosts) * sizeof(int));
  Next = malloc((NumPart + Nghosts) * sizeof(int));
  Tail = malloc((NumPart + Nghosts) * sizeof(int));
  GridNext = malloc((NumPart + Nghosts) * sizeof(int));

  if(!(GridNext))
    {
      printf("failed to allocate memory\n");
      endrun(1231);
    }

  for(i = 0; i < NumPart + Nghosts; i++)
    {
      Head[i] = Tail[i] = i;
      Len[i] = 1;
      Next[i] = -1;

      if(i < NumPart)
	GrIndex[i] = i + (((long long) ThisTask) << 48);
      else
	GrIndex[i] = ImportBuf[i - NumPart].GrIndex;
    }

  fof_course_binning();

  if(ThisTask == 0)
    {
      printf("Coarse binning done.\n");
      fflush(stdout);
    }

  fof_find_groups();

  fof_link_special();

  if(ThisTask == 0)
    {
      printf("Local groups found.\n");
      fflush(stdout);
    }

  fof_find_minids();

  if(ThisTask == 0)
    {
      printf("minids found.\n");
      fflush(stdout);
    }

  do
    {
      links = fof_link_accross();

      if(ThisTask == 0)
	printf("Made %d links accross domain boundaries...\n", links);
    }
  while(links > 0);


  if(ThisTask == 0)
    printf("Groups linked accross processor boundaries.\n");

  free(GridFlag);
  free(GridFirst);
  free(GridNext);
  free(Tail);
  free(Next);
  free(Len);


  fof_exchange_id_lists();

  fof_compile_catalogue();


  MPI_Allreduce(&Ngroups, &TotNgroups, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  MPI_Allgather(&Nids, 1, MPI_INT, local_toGo, 1, MPI_INT, MPI_COMM_WORLD);
  for(i = 0, TotNids = 0; i < NTask; i++)
    TotNids += local_toGo[i];

  if(TotNgroups > 0)
    MPI_Allreduce(&GroupLen[0], &largestgroup, 1, MPI_INT, MPI_MAX, MPI_COMM_WORLD);

  if(ThisTask == 0)
    {
      printf("\nTotal number of groups with at least %d particles: %d\n", GROUP_MIN_LEN, TotNgroups);
      if(TotNgroups > 0)
	{
	  printf("Largest group has %d particles.\n", largestgroup);
	  printf("Total number of particles in groups: %d%09d\n\n",
		 (int) (TotNids / 1000000000), (int) (TotNids % 1000000000));
	}
    }

#ifdef BLACK_HOLES
  if(num < 0)
    fof_make_black_holes();
#endif

  if(num >= 0)
    {
      fof_save_groups(num);
      
      if(ThisTask == 0)
	{
	  printf("Group catalogues saved.\n\n");
	  fflush(stdout);
	}
    }

  free(GroupIDs);
  free(GroupOffset);
  free(GroupLenType);
  free(GroupLen);
  free(local_ids);
  free(GrIndex);
  free(Head);
  free(ImportBuf);
  free(ExportBuf);
  free(import_offset);
  free(export_offset);
  free(toGo);
  free(local_toGo);

  free(fof_nearest_task);
  free(fof_nearest_index);
}



void fof_import_ghosts(void)
{
  int n, j, k;
  int level, sendTask, recvTask;
  MPI_Status status;

  if(ThisTask == 0)
    {
      printf("Start determining of ghosts....\n");
      fflush(stdout);
    }

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  for(n = 0; n < NumPart; n++)
    {
      for(j = 0; j < NTask; j++)
	Exportflag[j] = 0;

      if(P[n].Type == 1)
	{
	  ngb_treefind_flagexport(P[n].Pos, LinkL);
	}
      else
	{
	  if(fof_nearest_task[n] != ThisTask)
	    Exportflag[fof_nearest_task[n]] = 1;
	}

      for(j = 0; j < NTask; j++)
	if(Exportflag[j])
	  local_toGo[j] += 1;
    }

  MPI_Allgather(local_toGo, NTask, MPI_INT, toGo, NTask, MPI_INT, MPI_COMM_WORLD);

  for(j = 0, Nghosts = 0, Nexport = 0; j < NTask; j++)
    {
      Nghosts += toGo[j * NTask + ThisTask];
      Nexport += toGo[ThisTask * NTask + j];
    }

  printf("Task=%d needs to import %d ghost particles for FoF, and exports %d particles.\n", ThisTask, Nghosts,
	 Nexport);
  fflush(stdout);

  ExportBuf = malloc(1 + Nexport * sizeof(struct export_fof));
  ImportBuf = malloc(1 + Nghosts * sizeof(struct export_fof));

  if(!(ImportBuf) || !(ExportBuf))
    {
      printf("failed to allocate memory\n");
      endrun(12311);
    }

  export_offset[0] = 0;
  import_offset[0] = 0;

  for(n = 1; n < NTask; n++)
    {
      export_offset[n] = export_offset[n - 1] + toGo[ThisTask * NTask + (n - 1)];
      import_offset[n] = import_offset[n - 1] + toGo[(n - 1) * NTask + ThisTask];
    }


  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  for(n = 0; n < NumPart; n++)
    {
      for(j = 0; j < NTask; j++)
	Exportflag[j] = 0;

      if(P[n].Type == 1)
	{
	  ngb_treefind_flagexport(P[n].Pos, LinkL);
	}
      else
	{
	  if(fof_nearest_task[n] != ThisTask)
	    Exportflag[fof_nearest_task[n]] = 1;
	}


      for(j = 0; j < NTask; j++)
	if(Exportflag[j])
	  {
	    ExportBuf[export_offset[j] + local_toGo[j]].GrIndex = n + (((long long) ThisTask) << 48);
	    ExportBuf[export_offset[j] + local_toGo[j]].Type = P[n].Type;
	    ExportBuf[export_offset[j] + local_toGo[j]].NearestIndex = fof_nearest_index[n];
	    ExportBuf[export_offset[j] + local_toGo[j]].NearestTask = fof_nearest_task[n];

	    for(k = 0; k < 3; k++)
	      ExportBuf[export_offset[j] + local_toGo[j]].Pos[k] = P[n].Pos[k];

	    local_toGo[j] += 1;
	  }
    }


  for(level = 1; level < (1 << PTask); level++)
    {
      sendTask = ThisTask;
      recvTask = ThisTask ^ level;

      if(recvTask < NTask)
	MPI_Sendrecv(&ExportBuf[export_offset[recvTask]],
		     toGo[ThisTask * NTask + recvTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY,
		     &ImportBuf[import_offset[recvTask]],
		     toGo[recvTask * NTask + ThisTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
    }


  if(ThisTask == 0)
    {
      printf("exchange of ghosts done\n");
      fflush(stdout);
    }
}


void fof_course_binning(void)
{
  int i, j, k, n;
  double fac;
  double xmin[3], xmax[3], len;


  /* determine extension */
  for(j = 0; j < 3; j++)
    {
      xmin[j] = MAX_REAL_NUMBER;
      xmax[j] = -MAX_REAL_NUMBER;
    }

  for(i = 0; i < NumPart; i++)
    {
      for(j = 0; j < 3; j++)
	{
	  if(xmin[j] > P[i].Pos[j])
	    xmin[j] = P[i].Pos[j];

	  if(xmax[j] < P[i].Pos[j])
	    xmax[j] = P[i].Pos[j];
	}
    }

  for(i = 0; i < Nghosts; i++)
    {
      for(j = 0; j < 3; j++)
	{
	  if(ImportBuf[i].Pos[j] < xmin[j])
	    {
	      if((ImportBuf[i].Pos[j] + All.BoxSize - xmax[j]) < (xmin[j] - ImportBuf[i].Pos[j]))
		ImportBuf[i].Pos[j] += All.BoxSize;
	    }
	  else if(ImportBuf[i].Pos[j] > xmax[j])
	    {
	      if((xmin[j] - (ImportBuf[i].Pos[j] - All.BoxSize)) < (ImportBuf[i].Pos[j] - xmax[j]))
		ImportBuf[i].Pos[j] -= All.BoxSize;
	    }
	}
    }

  for(i = 0; i < Nghosts; i++)
    {
      for(j = 0; j < 3; j++)
	{
	  if(xmin[j] > ImportBuf[i].Pos[j])
	    xmin[j] = ImportBuf[i].Pos[j];

	  if(xmax[j] < ImportBuf[i].Pos[j])
	    xmax[j] = ImportBuf[i].Pos[j];
	}
    }

  len = 0;
  for(j = 0; j < 3; j++)
    if(xmax[j] - xmin[j] > len)
      len = xmax[j] - xmin[j];

  /* determine grid size */

  Grid = pow(NumPart, 1.0 / 3);

  /* we need to make sure that one mesh cell is at least of thickness LinkL */

  if(len / Grid < LinkL)
    Grid = len / LinkL;

  fac = Grid / len;


  GridFirst = malloc(Grid * Grid * Grid * sizeof(int));
  GridFlag = malloc(Grid * Grid * Grid * sizeof(char));

  if(!(GridFirst) || !(GridFlag))
    {
      printf("failed to allocate memory\n");
      endrun(112311);
    }

  for(n = 0; n < Grid * Grid * Grid; n++)
    {
      GridFirst[n] = -1;
      GridFlag[n] = 0;
    }

  for(n = 0; n < NumPart + Nghosts; n++)
    GridNext[n] = -1;


  for(n = 0; n < NumPart + Nghosts; n++)
    {
      if(n < NumPart)
	{
	  if(P[n].Type != 1)
	    continue;
	  i = (P[n].Pos[0] - xmin[0]) * fac;
	  j = (P[n].Pos[1] - xmin[1]) * fac;
	  k = (P[n].Pos[2] - xmin[2]) * fac;
	}
      else
	{
	  if(ImportBuf[n - NumPart].Type != 1)
	    continue;
	  i = (ImportBuf[n - NumPart].Pos[0] - xmin[0]) * fac;
	  j = (ImportBuf[n - NumPart].Pos[1] - xmin[1]) * fac;
	  k = (ImportBuf[n - NumPart].Pos[2] - xmin[2]) * fac;
	}

      if(i >= Grid)
	i = Grid - 1;
      if(j >= Grid)
	j = Grid - 1;
      if(k >= Grid)
	k = Grid - 1;

      if(GridFirst[(i * Grid + j) * Grid + k] >= 0)
	{
	  GridNext[n] = GridFirst[(i * Grid + j) * Grid + k];
	  GridFirst[(i * Grid + j) * Grid + k] = n;
	}
      else
	{
	  GridFirst[(i * Grid + j) * Grid + k] = n;
	}
    }
}



void fof_find_groups(void)
{
  int i, j, k;
  int p;

  if(ThisTask == 0)
    {
      printf("linking...");
      fflush(stdout);
    }

  for(i = Grid - 1; i >= 0; i--)
    for(j = Grid - 1; j >= 0; j--)
      for(k = Grid - 1; k >= 0; k--)
	{
	  if((p = GridFirst[(i * Grid + j) * Grid + k]) >= 0)
	    {
	      do
		{
		  fof_check_cell(p, i + 1, j, k);
		  fof_check_cell(p, i + 1, j + 1, k);
		  fof_check_cell(p, i + 1, j, k + 1);
		  fof_check_cell(p, i + 1, j + 1, k + 1);
		  fof_check_cell(p, i, j + 1, k);
		  fof_check_cell(p, i, j + 1, k + 1);
		  fof_check_cell(p, i, j, k + 1);

		  fof_check_cell(p, i, j, k);

		  fof_check_cell(p, i + 1, j, k - 1);
		  fof_check_cell(p, i + 1, j - 1, k);
		  fof_check_cell(p, i, j - 1, k + 1);
		  fof_check_cell(p, i - 1, j + 1, k + 1);
		  fof_check_cell(p, i - 1, j - 1, k + 1);
		  fof_check_cell(p, i + 1, j - 1, k + 1);
		}
	      while((p = GridNext[p]) >= 0);
	    }
	}
}



void fof_check_cell(int p, int i, int j, int k)
{
  double r2, s2, dx, dy, dz;
  int pp, ss;
  int s, flag;

  if(i < 0)
    i += Grid;
  if(j < 0)
    j += Grid;
  if(k < 0)
    k += Grid;

  if(i >= Grid)
    i -= Grid;
  if(j >= Grid)
    j -= Grid;
  if(k >= Grid)
    k -= Grid;


  if((s = GridFirst[(i * Grid + j) * Grid + k]) >= 0)
    {
      flag = Head[s];
      if(GridFlag[(i * Grid + j) * Grid + k])
	{
	  if(Head[s] == Head[p])
	    return;
	}
    }
  else
    flag = -1;

  while(s >= 0)
    {
      if(Head[s] != flag)
	flag = -1;

      if(Head[p] != Head[s])	/* only if not yet linked */
	{
	  if(p < NumPart)
	    {
	      dx = P[p].Pos[0];
	      dy = P[p].Pos[1];
	      dz = P[p].Pos[2];
	    }
	  else
	    {
	      dx = ImportBuf[p - NumPart].Pos[0];
	      dy = ImportBuf[p - NumPart].Pos[1];
	      dz = ImportBuf[p - NumPart].Pos[2];
	    }
	  if(s < NumPart)
	    {
	      dx -= P[s].Pos[0];
	      dy -= P[s].Pos[1];
	      dz -= P[s].Pos[2];
	    }
	  else
	    {
	      dx -= ImportBuf[s - NumPart].Pos[0];
	      dy -= ImportBuf[s - NumPart].Pos[1];
	      dz -= ImportBuf[s - NumPart].Pos[2];
	    }

	  dx = fof_periodic(dx);
	  dy = fof_periodic(dy);
	  dz = fof_periodic(dz);

	  r2 = dx * dx + dy * dy + dz * dz;

	  s2 = LinkL * LinkL;

	  if(r2 <= s2)
	    {
	      if(Len[Head[p]] > Len[Head[s]])	/* p group is longer */
		{
		  Next[Tail[Head[p]]] = Head[s];

		  Tail[Head[p]] = Tail[Head[s]];

		  Len[Head[p]] += Len[Head[s]];

		  ss = Head[s];
		  do
		    {
		      Head[ss] = Head[p];
		    }
		  while((ss = Next[ss]) >= 0);

		  flag = -1;
		}
	      else
		{
		  Next[Tail[Head[s]]] = Head[p];

		  Tail[Head[s]] = Tail[Head[p]];

		  Len[Head[s]] += Len[Head[p]];

		  pp = Head[p];
		  do
		    {
		      Head[pp] = Head[s];
		    }
		  while((pp = Next[pp]) >= 0);

		  flag = -1;
		}

	      if(GridFlag[(i * Grid + j) * Grid + k])
		return;
	    }
	}

      s = GridNext[s];
    }

  if(flag >= 0)
    GridFlag[(i * Grid + j) * Grid + k] = 1;
}


void fof_link_special(void)
{
  int n;

  for(n = 0; n < NumPart + Nghosts; n++)
    {
      if(n < NumPart)
	{
	  if(P[n].Type != 1)
	    if(fof_nearest_task[n] == ThisTask)
	      {
		fof_link_specialpair(n, fof_nearest_index[n]);
	      }
	}
      else
	{
	  if(ImportBuf[n - NumPart].Type != 1)
	    {
	      if(ImportBuf[n - NumPart].NearestTask == ThisTask)
		{
		  if(ImportBuf[n - NumPart].NearestIndex >= NumPart)
		    endrun(888);
		  fof_link_specialpair(n, ImportBuf[n - NumPart].NearestIndex);
		}
	    }
	}
    }
}

void fof_link_specialpair(int p, int s)
{
  int ss, pp;

  if(Head[p] != Head[s])	/* only if not yet linked */
    {
      if(Len[Head[p]] > Len[Head[s]])	/* p group is longer */
	{
	  Next[Tail[Head[p]]] = Head[s];

	  Tail[Head[p]] = Tail[Head[s]];

	  Len[Head[p]] += Len[Head[s]];

	  ss = Head[s];
	  do
	    {
	      Head[ss] = Head[p];
	    }
	  while((ss = Next[ss]) >= 0);
	}
      else
	{
	  Next[Tail[Head[s]]] = Head[p];

	  Tail[Head[s]] = Tail[Head[p]];

	  Len[Head[s]] += Len[Head[p]];

	  pp = Head[p];
	  do
	    {
	      Head[pp] = Head[s];
	    }
	  while((pp = Next[pp]) >= 0);
	}
    }
}








void fof_find_minids(void)
{
  int n, pp;
  long long minid, id;

  int grfound = 0;

  for(n = 0; n < NumPart + Nghosts; n++)
    {
      if(Head[n] == n)
	{
	  if(n < NumPart)
	    minid = n + (((long long) ThisTask) << 48);
	  else
	    minid = ImportBuf[n - NumPart].GrIndex;

	  pp = n;
	  do
	    {
	      if(pp < NumPart)
		id = pp + (((long long) ThisTask) << 48);
	      else
		id = ImportBuf[pp - NumPart].GrIndex;

	      if(minid > id)
		minid = id;
	    }
	  while((pp = Next[pp]) >= 0);

	  GrIndex[n] = minid;

	  grfound++;
	}
    }
}




int fof_link_accross(void)
{
  int n, j, links, tot_links;
  int level, sendTask, recvTask;
  long long headid;
  MPI_Status status;

  links = 0;

  for(n = 0; n < Nghosts; n++)
    ImportBuf[n].GrIndex = GrIndex[Head[n + NumPart]];

  for(level = 1; level < (1 << PTask); level++)
    {
      sendTask = ThisTask;
      recvTask = ThisTask ^ level;

      if(recvTask < NTask)
	MPI_Sendrecv(&ImportBuf[import_offset[recvTask]],
		     toGo[recvTask * NTask + ThisTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY,
		     &ExportBuf[export_offset[recvTask]],
		     toGo[ThisTask * NTask + recvTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
    }

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  for(n = 0; n < NumPart; n++)
    {
      for(j = 0; j < NTask; j++)
	Exportflag[j] = 0;

      if(P[n].Type == 1)
	{
	  ngb_treefind_flagexport(P[n].Pos, LinkL);
	}
      else
	{
	  if(fof_nearest_task[n] != ThisTask)
	    Exportflag[fof_nearest_task[n]] = 1;
	}

      for(j = 0; j < NTask; j++)
	if(Exportflag[j])
	  {
	    headid = ExportBuf[export_offset[j] + local_toGo[j]].GrIndex;

	    if(headid < GrIndex[Head[n]])
	      {
		GrIndex[Head[n]] = headid;
		links++;
	      }
	    else
	      headid = GrIndex[Head[n]];

	    ExportBuf[export_offset[j] + local_toGo[j]].GrIndex = headid;

	    local_toGo[j] += 1;
	  }
    }


  for(level = 1; level < (1 << PTask); level++)
    {
      sendTask = ThisTask;
      recvTask = ThisTask ^ level;

      if(recvTask < NTask)
	MPI_Sendrecv(&ExportBuf[export_offset[recvTask]],
		     toGo[ThisTask * NTask + recvTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY,
		     &ImportBuf[import_offset[recvTask]],
		     toGo[recvTask * NTask + ThisTask] * sizeof(struct export_fof),
		     MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
    }

  for(n = 0; n < Nghosts; n++)
    {
      headid = ImportBuf[n].GrIndex;

      if(headid < GrIndex[Head[n + NumPart]])
	{
	  GrIndex[Head[n + NumPart]] = headid;
	  links++;
	}
    }

  MPI_Allreduce(&links, &tot_links, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  return tot_links;
}




void fof_exchange_id_lists(void)
{
  int n;
  long long headid;
  int task, nsend, nget;
  int level, sendTask, recvTask;
  int j;

  MPI_Status status;

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  for(n = 0; n < NumPart; n++)
    {
      headid = GrIndex[Head[n]];
      task = (headid >> 48);

      if(task != ThisTask)
	local_toGo[task]++;
    }

  MPI_Allgather(local_toGo, NTask, MPI_INT, toGo, NTask, MPI_INT, MPI_COMM_WORLD);

  for(j = 0, nsend = nget = 0; j < NTask; j++)
    {
      nsend += toGo[ThisTask * NTask + j];
      nget += toGo[j * NTask + ThisTask];
    }

  printf("Task=%d needs to send %d particles for FoF-grouplist and gets %d particles.\n", ThisTask, nsend,
	 nget);
  fflush(stdout);

  Nlocal = NumPart - nsend + nget;

  local_ids = malloc(1 + Nlocal * sizeof(struct id_list));
  export_ids = malloc(1 + nsend * sizeof(struct id_list));

  if(!(export_ids) || !(local_ids))
    {
      printf("failed to allocate memory\n");
      endrun(152311);
    }

  import_offset[0] = NumPart - nsend;
  export_offset[0] = 0;

  for(n = 1; n < NTask; n++)
    {
      export_offset[n] = export_offset[n - 1] + toGo[ThisTask * NTask + (n - 1)];
      import_offset[n] = import_offset[n - 1] + toGo[(n - 1) * NTask + ThisTask];
    }

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  for(n = 0; n < NumPart; n++)
    {
      headid = GrIndex[Head[n]];
      task = headid >> 48;

      if(task != ThisTask)
	{
	  export_ids[export_offset[task] + local_toGo[task]].GrIndex = headid;
	  export_ids[export_offset[task] + local_toGo[task]].ID = P[n].ID;
	  export_ids[export_offset[task] + local_toGo[task]].Type = P[n].Type;
	  export_ids[export_offset[task] + local_toGo[task]].Task = ThisTask;
	  export_ids[export_offset[task] + local_toGo[task]].Index = n;
	  if(P[n].Type == 0)
	    export_ids[export_offset[task] + local_toGo[task]].Density = SphP[n].Density;
	}
      else
	{
	  local_ids[local_toGo[task]].GrIndex = headid;
	  local_ids[local_toGo[task]].ID = P[n].ID;
	  local_ids[local_toGo[task]].Type = P[n].Type;
	  local_ids[local_toGo[task]].Task = ThisTask;
	  local_ids[local_toGo[task]].Index = n;
	  if(P[n].Type == 0)
	    local_ids[local_toGo[task]].Density = SphP[n].Density;
	}

      local_toGo[task]++;
    }

  for(level = 1; level < (1 << PTask); level++)
    {
      sendTask = ThisTask;
      recvTask = ThisTask ^ level;

      if(recvTask < NTask)
	MPI_Sendrecv(&export_ids[export_offset[recvTask]],
		     toGo[ThisTask * NTask + recvTask] * sizeof(struct id_list),
		     MPI_BYTE, recvTask, TAG_ANY,
		     &local_ids[import_offset[recvTask]],
		     toGo[recvTask * NTask + ThisTask] * sizeof(struct id_list),
		     MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
    }

  free(export_ids);
}



int fof_grid_compare(const void *a, const void *b)
{
  if(((struct id_list *) a)->GrIndex > (((struct id_list *) b)->GrIndex))
    return -1;

  if(((struct id_list *) a)->GrIndex < (((struct id_list *) b)->GrIndex))
    return +1;

  if(((struct id_list *) a)->Type < (((struct id_list *) b)->Type))
    return -1;

  if(((struct id_list *) a)->Type > (((struct id_list *) b)->Type))
    return +1;


  if(((struct id_list *) a)->ID < (((struct id_list *) b)->ID))
    return -1;

  if(((struct id_list *) a)->ID > (((struct id_list *) b)->ID))
    return +1;

  return 0;
}




void fof_compile_catalogue(void)
{
  int i, j, k, start, groups;

  qsort(local_ids, Nlocal, sizeof(struct id_list), fof_grid_compare);

  Ngroups = 0;
  Nids = 0;
  groups = 0;

  for(i = 0, start = 0; i < Nlocal; i++)
    {
      if(local_ids[i].GrIndex != local_ids[start].GrIndex)
	{
	  for(j = start; j < i; j++)
	    local_ids[j].GrIndex = groups + (((long long) (i - start)) << 32);

	  if((i - start) >= GROUP_MIN_LEN)
	    {
	      Ngroups++;
	      Nids += (i - start);
	    }

	  groups++;
	  start = i;
	}
    }

  /* finish last group */
  for(j = start; j < Nlocal; j++)
    local_ids[j].GrIndex = groups + (((long long) (Nlocal - start)) << 32);

  if((Nlocal - start) >= GROUP_MIN_LEN)
    {
      Ngroups++;
      Nids += (Nlocal - start);
    }


  /* now sort the groups by size */
  qsort(local_ids, Nlocal, sizeof(struct id_list), fof_grid_compare);

  GroupLen = malloc(1 + Ngroups * sizeof(int));
  GroupLenType = malloc(6 * (1 + Ngroups * sizeof(int)));
  GroupOffset = malloc(1 + Ngroups * sizeof(int));
  GroupIDs = malloc(1 + Nids * sizeof(int));

  if(!(GroupLen) || !(GroupOffset) || !(GroupIDs))
    {
      printf("failed to allocate memory\n");
      endrun(1523711);
    }

  for(i = 0, j = 0, GroupOffset[0] = 0; i < Ngroups; i++)
    {
      GroupLen[i] = (local_ids[j].GrIndex >> 32);

      for(k = 0; k < 6; k++)
	GroupLenType[i * 6 + k] = 0;

      for(k = 0; k < GroupLen[i]; k++)
	GroupLenType[i * 6 + local_ids[j + k].Type]++;

      j += GroupLen[i];
      if(i > 0)
	GroupOffset[i] = GroupOffset[i - 1] + GroupLen[i - 1];
    }

  for(i = 0; i < Nids; i++)
    GroupIDs[i] = local_ids[i].ID;
}


#ifdef BLACK_HOLES

void fof_make_black_holes(void)
{
  int i, j, k, n, index, nblackhole, ntot, task;
  int nsend, nget, nlocal, sendTask, recvTask, level;
  int *local_indices, *export_indices;
  double maxdens;
  MPI_Status status;

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;

  nblackhole = 0;

  for(i = 0; i < Ngroups; i++)
    {
      if(GroupLenType[i * 6 + 1] >= MIN_HALO_SIZE_FOR_BLACK_HOLE)
	if(GroupLenType[i * 6 + 5] == 0)
	  {
	    maxdens = 0;
	    index = -1;
	    for(k = 0; k < GroupLen[i]; k++)
	      if(local_ids[GroupOffset[i] + k].Type == 0)
		{
		  if(local_ids[GroupOffset[i] + k].Density > maxdens)
		    {
#ifdef WINDS
			if(SphP[i].DelayTime == 0)
			{
#endif
			    maxdens = local_ids[GroupOffset[i] + k].Density;
			    index = GroupOffset[i] + k;
			}
		    }

		}
	    if(index >= 0)
	      {
		nblackhole++;
		if(local_ids[index].Task != ThisTask)
		  local_toGo[local_ids[index].Task]++;
	      }
	  }
    }


  MPI_Allgather(local_toGo, NTask, MPI_INT, toGo, NTask, MPI_INT, MPI_COMM_WORLD);

  for(j = 0, nsend = nget = 0; j < NTask; j++)
    {
      nsend += toGo[ThisTask * NTask + j];
      nget += toGo[j * NTask + ThisTask];
    }

  nlocal = nblackhole - nsend + nget;

  local_indices = malloc(nlocal * sizeof(int));
  export_indices = malloc(nsend * sizeof(int));

  import_offset[0] = nblackhole - nsend;
  export_offset[0] = 0;

  for(n = 1; n < NTask; n++)
    {
      export_offset[n] = export_offset[n - 1] + toGo[ThisTask * NTask + (n - 1)];
      import_offset[n] = import_offset[n - 1] + toGo[(n - 1) * NTask + ThisTask];
    }

  for(n = 0; n < NTask; n++)
    local_toGo[n] = 0;


  for(i = 0; i < Ngroups; i++)
    {
      if(GroupLenType[i * 6 + 1] >= MIN_HALO_SIZE_FOR_BLACK_HOLE)
	if(GroupLenType[i * 6 + 5] == 0)
	  {
	    maxdens = 0;
	    index = -1;
	    for(k = 0; k < GroupLen[i]; k++)
	      if(local_ids[GroupOffset[i] + k].Type == 0)
		{
		  if(local_ids[GroupOffset[i] + k].Density > maxdens)
		    {
		      maxdens = local_ids[GroupOffset[i] + k].Density;
		      index = GroupOffset[i] + k;
		    }

		}
	    if(index >= 0)
	      {
		if(local_ids[index].Task != ThisTask)
		  {
		    task = local_ids[index].Task;
		    export_indices[export_offset[task] + local_toGo[task]] = local_ids[index].Index;
		  }
		else
		  {
		    local_indices[local_toGo[ThisTask]] = local_ids[index].Index;
		  }
		local_toGo[local_ids[index].Task]++;


		GroupLenType[i * 6 + 1]--;
		GroupLenType[i * 6 + 5]++;
	      }
	  }
    }


  for(level = 1; level < (1 << PTask); level++)
    {
      sendTask = ThisTask;
      recvTask = ThisTask ^ level;

      if(recvTask < NTask)
	MPI_Sendrecv(&export_indices[export_offset[recvTask]],
		     toGo[ThisTask * NTask + recvTask] * sizeof(int),
		     MPI_BYTE, recvTask, TAG_ANY,
		     &local_indices[import_offset[recvTask]],
		     toGo[recvTask * NTask + ThisTask] * sizeof(int),
		     MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
    }

  MPI_Allreduce(&nblackhole, &ntot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  if(ThisTask == 0)
    {
      printf("\nMaking %d new black hole particles\n\n", ntot);
      fflush(stdout);
    }

  for(n = 0; n < nlocal; n++)
    {
      if(P[local_indices[n]].Type != 0)
	endrun(7772);

      P[local_indices[n]].Type = 5;	/* make it a black hole particle */
      P[local_indices[n]].StellarAge = All.Time;

      Stars_converted++;
    }

  All.TotN_gas -= ntot;

  free(export_indices);
  free(local_indices);
}

#endif






void fof_save_groups(int num)
{
  int nprocgroup, masterTask, groupTask;

  if(NTask < All.NumFilesWrittenInParallel)
    {
      printf
	("Fatal error.\nNumber of processors must be a smaller or equal than `NumFilesWrittenInParallel'.\n");
      endrun(241931);
    }

  nprocgroup = NTask / All.NumFilesWrittenInParallel;

  if((NTask % All.NumFilesWrittenInParallel))
    nprocgroup++;

  masterTask = (ThisTask / nprocgroup) * nprocgroup;

  for(groupTask = 0; groupTask < nprocgroup; groupTask++)
    {
      if(ThisTask == (masterTask + groupTask))	/* ok, it's this processor's turn */
	fof_save_local_catalogue(num);

      MPI_Barrier(MPI_COMM_WORLD);	/* wait inside the group */
    }
}




void fof_save_local_catalogue(int num)
{
  FILE *fd;
  char buf[500];

  sprintf(buf, "%s/%s_%03d.%d", All.OutputDir, "group_tab", num, ThisTask);

  if(!(fd = fopen(buf, "w")))
    {
      printf("can't open file `%s`\n", buf);
      endrun(1183);
    }

  fwrite(&Ngroups, sizeof(int), 1, fd);
  fwrite(&Nids, sizeof(int), 1, fd);
  fwrite(&TotNgroups, sizeof(int), 1, fd);
  fwrite(&NTask, sizeof(int), 1, fd);
  fwrite(GroupLen, sizeof(int), Ngroups, fd);
  fwrite(GroupOffset, sizeof(int), Ngroups, fd);
  fwrite(GroupLenType,  6*sizeof(int), Ngroups, fd);
  fclose(fd);

  sprintf(buf, "%s/%s_%03d.%d", All.OutputDir, "group_ids", num, ThisTask);

  if(!(fd = fopen(buf, "w")))
    {
      printf("can't open file `%s`\n", buf);
      endrun(1184);
    }

  fwrite(&Ngroups, sizeof(int), 1, fd);
  fwrite(&Nids, sizeof(int), 1, fd);
  fwrite(&TotNgroups, sizeof(int), 1, fd);
  fwrite(&NTask, sizeof(int), 1, fd);
  fwrite(GroupIDs, sizeof(int), Nids, fd);
  fclose(fd);
}





double fof_periodic(double x)
{
  if(x >= 0.5 * All.BoxSize)
    x -= All.BoxSize;

  if(x < -0.5 * All.BoxSize)
    x += All.BoxSize;

  return x;
}


double fof_periodic_wrap(double x)
{
  while(x >= All.BoxSize)
    x -= All.BoxSize;

  while(x < 0)
    x += All.BoxSize;

  return x;
}


void fof_find_nearest_dmparticle(void)
{
  int i, j, n, count, ntot, ntotleft, ndonetot, level, maxfill;
  int nexport, ndone, ngrp, sendTask, recvTask, place, source, npleft, iter;
  int *noffset, *nbuffer, *nsend_local, *nsend;
  MPI_Status status;

  fof_nearest_distance = malloc(sizeof(float) * NumPart);
  fof_nearest_hsml = malloc(sizeof(float) * NumPart);

  noffset = malloc(sizeof(int) * NTask);	/* offsets of bunches in common list */
  nbuffer = malloc(sizeof(int) * NTask);
  nsend_local = malloc(sizeof(int) * NTask);
  nsend = malloc(sizeof(int) * NTask * NTask);


  if(ThisTask == 0)
    {
      printf("\nStart finding nearest dm-particle\n");
      fflush(stdout);
    }


  for(n = 0, count = 0; n < NumPart; n++)
    {
      if(P[n].Type != 1)
	{
	  fof_nearest_index[n] = -1;
	  fof_nearest_task[n] = -1;
	  fof_nearest_distance[n] = 1.0e30;
	  if(P[n].Type == 0)
	    fof_nearest_hsml[n] = PPP[n].Hsml;
	  else
	    fof_nearest_hsml[n] = 0.1 * LinkL;

	  count++;
	}
    }

  MPI_Allreduce(&count, &ntot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

  iter = 0;

  /* we will repeat the whole thing for those particles where we didn't find enough neighbours */
  do
    {
      i = 0;			/* beginn with this index */
      ntotleft = ntot;		/* particles left for all tasks together */

      while(ntotleft > 0)
	{
	  for(j = 0; j < NTask; j++)
	    nsend_local[j] = 0;

	  /* do local particles and prepare export list */

	  for(nexport = 0, ndone = 0; i < NumPart && nexport < All.BunchSizeFoF - NTask; i++)
	    if(P[i].Type != 1 && fof_nearest_task[i] == -1)
	      {
		ndone++;

		for(j = 0; j < NTask; j++)
		  Exportflag[j] = 0;

		fof_find_nearest_dmparticle_evaluate(i, 0);

		for(j = 0; j < NTask; j++)
		  {
		    if(Exportflag[j])
		      {
			FoFDataIn[nexport].Pos[0] = P[i].Pos[0];
			FoFDataIn[nexport].Pos[1] = P[i].Pos[1];
			FoFDataIn[nexport].Pos[2] = P[i].Pos[2];
			FoFDataIn[nexport].Hsml = fof_nearest_hsml[i];
			FoFDataIn[nexport].Index = i;
			FoFDataIn[nexport].Task = j;
			nexport++;
			nsend_local[j]++;
		      }
		  }
	      }

	  qsort(FoFDataIn, nexport, sizeof(struct fofdata_in), fof_compare_key);

	  for(j = 1, noffset[0] = 0; j < NTask; j++)
	    noffset[j] = noffset[j - 1] + nsend_local[j - 1];

	  MPI_Allgather(nsend_local, NTask, MPI_INT, nsend, NTask, MPI_INT, MPI_COMM_WORLD);

	  /* now do the particles that need to be exported */

	  for(level = 1; level < (1 << PTask); level++)
	    {
	      for(j = 0; j < NTask; j++)
		nbuffer[j] = 0;
	      for(ngrp = level; ngrp < (1 << PTask); ngrp++)
		{
		  maxfill = 0;
		  for(j = 0; j < NTask; j++)
		    {
		      if((j ^ ngrp) < NTask)
			if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
			  maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		    }
		  if(maxfill >= All.BunchSizeFoF)
		    break;

		  sendTask = ThisTask;
		  recvTask = ThisTask ^ ngrp;

		  if(recvTask < NTask)
		    {
		      if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
			{
			  /* get the particles */
			  MPI_Sendrecv(&FoFDataIn[noffset[recvTask]],
				       nsend_local[recvTask] * sizeof(struct fofdata_in), MPI_BYTE,
				       recvTask, TAG_ANY,
				       &FoFDataGet[nbuffer[ThisTask]],
				       nsend[recvTask * NTask + ThisTask] * sizeof(struct fofdata_in),
				       MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
			}
		    }

		  for(j = 0; j < NTask; j++)
		    if((j ^ ngrp) < NTask)
		      nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
		}


	      for(j = 0; j < nbuffer[ThisTask]; j++)
		fof_find_nearest_dmparticle_evaluate(j, 1);

	      /* get the result */
	      for(j = 0; j < NTask; j++)
		nbuffer[j] = 0;
	      for(ngrp = level; ngrp < (1 << PTask); ngrp++)
		{
		  maxfill = 0;
		  for(j = 0; j < NTask; j++)
		    {
		      if((j ^ ngrp) < NTask)
			if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
			  maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		    }
		  if(maxfill >= All.BunchSizeFoF)
		    break;

		  sendTask = ThisTask;
		  recvTask = ThisTask ^ ngrp;

		  if(recvTask < NTask)
		    {
		      if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
			{
			  /* send the results */
			  MPI_Sendrecv(&FoFDataResult[nbuffer[ThisTask]],
				       nsend[recvTask * NTask + ThisTask] * sizeof(struct fofdata_out),
				       MPI_BYTE, recvTask, TAG_ANY,
				       &FoFDataPartialResult[noffset[recvTask]],
				       nsend_local[recvTask] * sizeof(struct fofdata_out),
				       MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);

			  /* add the result to the particles */
			  for(j = 0; j < nsend_local[recvTask]; j++)
			    {
			      source = j + noffset[recvTask];
			      place = FoFDataIn[source].Index;

			      if(FoFDataPartialResult[source].Index >= 0)
				{
				  if(FoFDataPartialResult[source].Distance < fof_nearest_distance[place])
				    {
				      fof_nearest_distance[place] = FoFDataPartialResult[source].Distance;
				      fof_nearest_index[place] = FoFDataPartialResult[source].Index;
				      fof_nearest_task[place] = recvTask;
				    }
				}
			    }
			}
		    }

		  for(j = 0; j < NTask; j++)
		    if((j ^ ngrp) < NTask)
		      nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
		}

	      level = ngrp - 1;
	    }

	  MPI_Allreduce(&ndone, &ndonetot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

	  ntotleft -= ndonetot;
	}

      /* do final operations on results */
      for(i = 0, npleft = 0; i < NumPart; i++)
	{
	  if(P[i].Type != 1)
	    {
	      if(fof_nearest_task[i] == -1)
		{
		  /* need to redo this particle */
		  npleft++;
		  fof_nearest_hsml[i] *= 2.0;

		  if(iter >= MAXITER - 10)
		    {
		      printf("i=%d task=%d ID=%d Hsml=%g  pos=(%g|%g|%g)\n",
			     i, ThisTask, P[i].ID, fof_nearest_hsml[i],
			     P[i].Pos[0], P[i].Pos[1], P[i].Pos[2]);
		      fflush(stdout);
		    }
		}
	    }
	}

      MPI_Allreduce(&npleft, &ntot, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);

      if(ntot > 0)
	{
	  iter++;

	  if(iter > 0 && ThisTask == 0)
	    {
	      printf("fof-nearest iteration %d: need to repeat for %d particles.\n", iter, ntot);
	      fflush(stdout);
	    }

	  if(iter > MAXITER)
	    {
	      printf("failed to converge in fof-nearest\n");
	      fflush(stdout);
	      endrun(1159);
	    }
	}
    }
  while(ntot > 0);


  free(nsend);
  free(nsend_local);
  free(nbuffer);
  free(noffset);

  free(fof_nearest_hsml);
  free(fof_nearest_distance);


  if(ThisTask == 0)
    {
      printf("\ndone finding nearest dm-particle\n");
      fflush(stdout);
    }
}


void fof_find_nearest_dmparticle_evaluate(int target, int mode)
{
  int j, n, index;
  int startnode, numngb_inbox;
  double h, r2max;
  double dx, dy, dz, r2;
  FLOAT pos[3];

#ifdef PERIODIC
  double boxSize, boxHalf;

  boxSize = All.BoxSize;
  boxHalf = 0.5 * All.BoxSize;
#endif

  if(mode == 0)
    {
      pos[0] = P[target].Pos[0];
      pos[1] = P[target].Pos[1];
      pos[2] = P[target].Pos[2];
      h = fof_nearest_hsml[target];
    }
  else
    {
      pos[0] = FoFDataGet[target].Pos[0];
      pos[1] = FoFDataGet[target].Pos[1];
      pos[2] = FoFDataGet[target].Pos[2];
      h = FoFDataGet[target].Hsml;
    }

  startnode = All.MaxPart;

  index = -1;
  r2max = 1.0e30;

  do
    {
      numngb_inbox = ngb_treefind_darkmatter(&pos[0], h, &startnode);

      for(n = 0; n < numngb_inbox; n++)
	{
	  j = Ngblist[n];

	  dx = pos[0] - P[j].Pos[0];
	  dy = pos[1] - P[j].Pos[1];
	  dz = pos[2] - P[j].Pos[2];

#ifdef PERIODIC			/*  now find the closest image in the given box size  */
#ifdef LONGBOX
	  if(dx > boxHalf * LONGBOX)
	    dx -= boxSize * LONGBOX;
	  if(dx < -boxHalf * LONGBOX)
	    dx += boxSize * LONGBOX;
#else
	  if(dx > boxHalf)
	    dx -= boxSize;
	  if(dx < -boxHalf)
	    dx += boxSize;
#endif
	  if(dy > boxHalf)
	    dy -= boxSize;
	  if(dy < -boxHalf)
	    dy += boxSize;
	  if(dz > boxHalf)
	    dz -= boxSize;
	  if(dz < -boxHalf)
	    dz += boxSize;
#endif
	  r2 = dx * dx + dy * dy + dz * dz;

	  if(r2 < r2max)
	    {
	      index = j;
	      r2max = r2;
	    }
	}
    }
  while(startnode >= 0);


  if(mode == 0)
    {
      if(index >= 0)
	{
	  fof_nearest_index[target] = index;
	  fof_nearest_task[target] = ThisTask;
	  fof_nearest_distance[target] = sqrt(r2max);
	}
    }
  else
    {
      if(index >= 0)
	{
	  FoFDataResult[target].Index = index;
	  FoFDataResult[target].Distance = sqrt(r2max);
	}
      else
	FoFDataResult[target].Index = -1;
    }
}




/*! This routine is a comparison kernel used in a sort routine to group
 *  particles that are exported to the same processor.
 */
int fof_compare_key(const void *a, const void *b)
{
  if(((struct fofdata_in *) a)->Task < (((struct fofdata_in *) b)->Task))
    return -1;

  if(((struct fofdata_in *) a)->Task > (((struct fofdata_in *) b)->Task))
    return +1;

  return 0;
}



#endif
