#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <mpi.h>
#include <gsl/gsl_math.h>
#include "allvars.h"
#include "proto.h"

/*! \file hydra.c
 *  \brief Computation of SPH forces and rate of entropy generation
 *
 *  This file contains the "second SPH loop", where the SPH forces are
 *  computed, and where the rate of change of entropy due to the shock heating
 *  (via artificial viscosity) is computed.
 */



static double hubble_a, atime, hubble_a2, fac_mu, fac_vsic_fix, a3inv, fac_egy;

#ifdef PERIODIC
static double boxSize, boxHalf;
#endif

#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
static double u1;
#endif

/*! This function is the driver routine for the calculation of hydrodynamical
 *  force and rate of change of entropy due to shock heating for all active
 *  particles .
 */
void hydro_force(void)
{
  long long ntot, ntotleft;
  int i, j, k, n, ngrp, maxfill, source, ndone;
  int *nbuffer, *noffset, *nsend_local, *nsend, *numlist, *ndonelist;
  int level, sendTask, recvTask;
  int nexport, place;
  double soundspeed_i;
  double tstart, tend;
  double sumt, sumcomm;
  double timecomp = 0, timecommsumm = 0, timeimbalance = 0, sumimbalance;
  MPI_Status status;

#ifdef CONDUCTION
  double rEntropyFlow, dt, fac;
#endif

#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
  /* mean particle placing */
  u1 = pow(4. * M_PI / 3 / All.DesNumNgb, 1. / 3.);
#endif

#ifdef PERIODIC
  boxSize = All.BoxSize;
  boxHalf = 0.5 * All.BoxSize;
#endif

  if(All.ComovingIntegrationOn)
    {
      /* Factors for comoving integration of hydro */
      hubble_a = All.Omega0 / (All.Time * All.Time * All.Time)
	+ (1 - All.Omega0 - All.OmegaLambda) / (All.Time * All.Time)
#ifdef DARKENERGY
	+ DarkEnergy_a(All.Time);
#else
	+ All.OmegaLambda;
#endif
      hubble_a = All.Hubble * sqrt(hubble_a);
      hubble_a2 = All.Time * All.Time * hubble_a;

      fac_mu = pow(All.Time, 3 * (GAMMA - 1) / 2) / All.Time;

      fac_egy = pow(All.Time, 3 * (GAMMA - 1));

      fac_vsic_fix = hubble_a * pow(All.Time, 3 * GAMMA_MINUS1);

      a3inv = 1 / (All.Time * All.Time * All.Time);
      atime = All.Time;
    }
  else
    hubble_a = hubble_a2 = atime = fac_mu = fac_vsic_fix = a3inv = fac_egy = 1.0;


  /* `NumSphUpdate' gives the number of particles on this processor that want a force update */
  for(n = 0, NumSphUpdate = 0; n < N_gas; n++)
    {
#ifdef CONDUCTION
      SphP[n].CondEnergyChange = 0.0;
#endif

#ifdef SFR
      if(P[n].Type == 0)
#endif
	if(P[n].Ti_endstep == All.Ti_Current)
	  NumSphUpdate++;
    }

  numlist = malloc(NTask * sizeof(int) * NTask);
  MPI_Allgather(&NumSphUpdate, 1, MPI_INT, numlist, 1, MPI_INT, MPI_COMM_WORLD);
  for(i = 0, ntot = 0; i < NTask; i++)
    ntot += numlist[i];
  free(numlist);


  noffset = malloc(sizeof(int) * NTask);	/* offsets of bunches in common list */
  nbuffer = malloc(sizeof(int) * NTask);
  nsend_local = malloc(sizeof(int) * NTask);
  nsend = malloc(sizeof(int) * NTask * NTask);
  ndonelist = malloc(sizeof(int) * NTask);


  i = 0;			/* first particle for this task */
  ntotleft = ntot;		/* particles left for all tasks together */

  while(ntotleft > 0)
    {
      for(j = 0; j < NTask; j++)
	nsend_local[j] = 0;

      /* do local particles and prepare export list */
      tstart = second();
      for(nexport = 0, ndone = 0; i < N_gas && nexport < All.BunchSizeHydro - NTask; i++)
#ifdef SFR
	if(P[i].Type == 0)
#endif
	  if(P[i].Ti_endstep == All.Ti_Current)
	    {
	      ndone++;

	      for(j = 0; j < NTask; j++)
		Exportflag[j] = 0;

	      hydro_evaluate(i, 0);

	      for(j = 0; j < NTask; j++)
		{
		  if(Exportflag[j])
		    {
		      for(k = 0; k < 3; k++)
			{
			  HydroDataIn[nexport].Pos[k] = P[i].Pos[k];
			  HydroDataIn[nexport].Vel[k] = SphP[i].VelPred[k];
			}
		      HydroDataIn[nexport].Hsml = PPP[i].Hsml;
		      HydroDataIn[nexport].Mass = P[i].Mass;
#ifndef NOGRADHSML
		      HydroDataIn[nexport].DhsmlDensityFactor = SphP[i].DhsmlDensityFactor;
#endif
		      HydroDataIn[nexport].Density = SphP[i].Density;
		      HydroDataIn[nexport].Pressure = SphP[i].Pressure;
		      HydroDataIn[nexport].Timestep = P[i].Ti_endstep - P[i].Ti_begstep;

		      /* calculation of F1 */
		      soundspeed_i = sqrt(GAMMA * SphP[i].Pressure / SphP[i].Density);
		      HydroDataIn[nexport].F1 = fabs(SphP[i].DivVel) /
			(fabs(SphP[i].DivVel) + SphP[i].CurlVel +
			 0.0001 * soundspeed_i / PPP[i].Hsml / fac_mu);
#ifdef MAGNETIC
		      for(k = 0; k < 3; k++)
			HydroDataIn[nexport].BPred[k] = SphP[i].BPred[k];
#endif

#ifdef CONDUCTION
		      HydroDataIn[nexport].SmoothedEntr = SphP[i].SmoothedEntr;
		      HydroDataIn[nexport].Entropy = SphP[i].Entropy;
#ifdef CONDUCTION_SATURATION
		      HydroDataIn[nexport].GradEntr[0] = SphP[i].GradEntr[0];
		      HydroDataIn[nexport].GradEntr[1] = SphP[i].GradEntr[1];
		      HydroDataIn[nexport].GradEntr[2] = SphP[i].GradEntr[2];
#endif
#endif

#ifdef REDUCEVISC
		      HydroDataIn[nexport].alpha = SphP[i].alpha;
#endif

#ifdef SFR_DECOUPLING
		      HydroDataIn[nexport].DensityOld = SphP[i].DensityOld;
		      HydroDataIn[nexport].Entropy = SphP[i].Entropy;
#endif


#ifdef PARTICLE_DEBUG
		      HydroDataIn[nexport].ID = P[i].ID;
#endif
		      HydroDataIn[nexport].Index = i;
		      HydroDataIn[nexport].Task = j;
		      nexport++;
		      nsend_local[j]++;
		    }
		}
	    }
      tend = second();
      timecomp += timediff(tstart, tend);

      qsort(HydroDataIn, nexport, sizeof(struct hydrodata_in), hydro_compare_key);

      for(j = 1, noffset[0] = 0; j < NTask; j++)
	noffset[j] = noffset[j - 1] + nsend_local[j - 1];

      tstart = second();

      MPI_Allgather(nsend_local, NTask, MPI_INT, nsend, NTask, MPI_INT, MPI_COMM_WORLD);

      tend = second();
      timeimbalance += timediff(tstart, tend);



      /* now do the particles that need to be exported */

      for(level = 1; level < (1 << PTask); level++)
	{
	  tstart = second();
	  for(j = 0; j < NTask; j++)
	    nbuffer[j] = 0;
	  for(ngrp = level; ngrp < (1 << PTask); ngrp++)
	    {
	      maxfill = 0;
	      for(j = 0; j < NTask; j++)
		{
		  if((j ^ ngrp) < NTask)
		    if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
		      maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		}
	      if(maxfill >= All.BunchSizeHydro)
		break;

	      sendTask = ThisTask;
	      recvTask = ThisTask ^ ngrp;

	      if(recvTask < NTask)
		{
		  if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
		    {
		      /* get the particles */
		      MPI_Sendrecv(&HydroDataIn[noffset[recvTask]],
				   nsend_local[recvTask] * sizeof(struct hydrodata_in), MPI_BYTE,
				   recvTask, TAG_ANY,
				   &HydroDataGet[nbuffer[ThisTask]],
				   nsend[recvTask * NTask + ThisTask] * sizeof(struct hydrodata_in), MPI_BYTE,
				   recvTask, TAG_ANY, MPI_COMM_WORLD, &status);
		    }
		}

	      for(j = 0; j < NTask; j++)
		if((j ^ ngrp) < NTask)
		  nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
	    }
	  tend = second();
	  timecommsumm += timediff(tstart, tend);

	  /* now do the imported particles */
	  tstart = second();
	  for(j = 0; j < nbuffer[ThisTask]; j++)
	    {
	      hydro_evaluate(j, 1);
	    }
	  tend = second();
	  timecomp += timediff(tstart, tend);

	  /* do a block to measure imbalance */
	  tstart = second();
	  MPI_Barrier(MPI_COMM_WORLD);
	  tend = second();
	  timeimbalance += timediff(tstart, tend);

	  /* get the result */
	  tstart = second();
	  for(j = 0; j < NTask; j++)
	    nbuffer[j] = 0;
	  for(ngrp = level; ngrp < (1 << PTask); ngrp++)
	    {
	      maxfill = 0;
	      for(j = 0; j < NTask; j++)
		{
		  if((j ^ ngrp) < NTask)
		    if(maxfill < nbuffer[j] + nsend[(j ^ ngrp) * NTask + j])
		      maxfill = nbuffer[j] + nsend[(j ^ ngrp) * NTask + j];
		}
	      if(maxfill >= All.BunchSizeHydro)
		break;

	      sendTask = ThisTask;
	      recvTask = ThisTask ^ ngrp;

	      if(recvTask < NTask)
		{
		  if(nsend[ThisTask * NTask + recvTask] > 0 || nsend[recvTask * NTask + ThisTask] > 0)
		    {
		      /* send the results */
		      MPI_Sendrecv(&HydroDataResult[nbuffer[ThisTask]],
				   nsend[recvTask * NTask + ThisTask] * sizeof(struct hydrodata_out),
				   MPI_BYTE, recvTask, TAG_ANY,
				   &HydroDataPartialResult[noffset[recvTask]],
				   nsend_local[recvTask] * sizeof(struct hydrodata_out),
				   MPI_BYTE, recvTask, TAG_ANY, MPI_COMM_WORLD, &status);

		      /* add the result to the particles */
		      for(j = 0; j < nsend_local[recvTask]; j++)
			{
			  source = j + noffset[recvTask];
			  place = HydroDataIn[source].Index;

			  for(k = 0; k < 3; k++)
			    SphP[place].HydroAccel[k] += HydroDataPartialResult[source].Acc[k];

			  SphP[place].DtEntropy += HydroDataPartialResult[source].DtEntropy;

			  if(SphP[place].MaxMu < HydroDataPartialResult[source].MaxMu)
			    SphP[place].MaxMu = HydroDataPartialResult[source].MaxMu;
#ifdef CONDUCTION
			  SphP[place].CondEnergyChange += HydroDataPartialResult[source].CondEnergyChange;
#ifdef OUTPUTCOOLRATE
			  SphP[place].CondRate += HydroDataPartialResult[source].CondRate;
#endif
#endif
#ifdef MAGNETIC
			  for(k = 0; k < 3; k++)
			    SphP[place].DtB[k] += HydroDataPartialResult[source].DtB[k];
#ifdef TRACEDIVB
			  SphP[place].divB += HydroDataPartialResult[source].divB;
#endif
#endif
			}
		    }
		}

	      for(j = 0; j < NTask; j++)
		if((j ^ ngrp) < NTask)
		  nbuffer[j] += nsend[(j ^ ngrp) * NTask + j];
	    }
	  tend = second();
	  timecommsumm += timediff(tstart, tend);

	  level = ngrp - 1;
	}

      MPI_Allgather(&ndone, 1, MPI_INT, ndonelist, 1, MPI_INT, MPI_COMM_WORLD);
      for(j = 0; j < NTask; j++)
	ntotleft -= ndonelist[j];
    }

  free(ndonelist);
  free(nsend);
  free(nsend_local);
  free(nbuffer);
  free(noffset);



  /* do final operations on results */
  tstart = second();

  for(i = 0; i < N_gas; i++)
#ifdef SFR
    if(P[i].Type == 0)
#endif
      if(P[i].Ti_endstep == All.Ti_Current)
	{

	  SphP[i].DtEntropy *= GAMMA_MINUS1 / (hubble_a2 * pow(SphP[i].Density, GAMMA_MINUS1));

#ifdef MAGNETIC
	  /* take care of cosmological dilution */
	  if(All.ComovingIntegrationOn)
	    for(k = 0; k < 3; k++)
	      SphP[i].DtB[k] -= 2.0 * SphP[i].BPred[k];
#endif

#ifdef WINDS
	  /* if we have winds, we decouple particles briefly if delaytime>0 */

	  if(SphP[i].DelayTime > 0)
	    {
	      for(k = 0; k < 3; k++)
		SphP[i].HydroAccel[k] = 0;

	      SphP[i].DtEntropy = 0;
	      SphP[i].MaxMu = 0;
	    }
#endif
	}


#ifdef CONDUCTION		/* note: this needs to be done for all particles, not just active ones */
  for(i = 0; i < N_gas; i++)
#ifdef SFR
    if(P[i].Type == 0)
#endif
      {
	if(All.ComovingIntegrationOn)
	  SphP[i].CondEnergyChange *= All.Time / hubble_a;

	if(P[i].Ti_endstep > 0)
	  {
	    rEntropyFlow =
	      SphP[i].CondEnergyChange / P[i].Mass * GAMMA_MINUS1 * pow(SphP[i].Density * a3inv,
									-GAMMA_MINUS1);
	    if(fabs(rEntropyFlow) > 0.5 * SphP[i].Entropy)
	      {
		printf
		  ("NOTE: step=%d task=%d large entropy change for particle i=%d (ID=%d) due to conduction: A=%g  dA=%g rho=%g dt=%g\n",
		   All.NumCurrentTiStep, ThisTask, i, (int) P[i].ID, SphP[i].Entropy, rEntropyFlow,
		   SphP[i].Density, (P[i].Ti_endstep - P[i].Ti_begstep) * All.Timebase_interval);
		fflush(stdout);

		fac = 0.5 * SphP[i].Entropy / fabs(rEntropyFlow);

		rEntropyFlow *= fac;
		SphP[i].CondEnergyChange *= fac;
	      }

	    SphP[i].Entropy += rEntropyFlow;

	    if(P[i].Ti_endstep == All.Ti_Current)
	      {
		dt = (P[i].Ti_endstep - P[i].Ti_begstep) * All.Timebase_interval;
		SphP[i].CondEnergyChange /= (P[i].Mass * dt);
	      }
	  }
	else
	  {
	    SphP[i].CondEnergyChange /= P[i].Mass;
	  }
      }
#endif


  tend = second();
  timecomp += timediff(tstart, tend);

  /* collect some timing information */

  MPI_Reduce(&timecomp, &sumt, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&timecommsumm, &sumcomm, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  MPI_Reduce(&timeimbalance, &sumimbalance, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);

  if(ThisTask == 0)
    {
      All.CPU_HydCompWalk += sumt / NTask;
      All.CPU_HydCommSumm += sumcomm / NTask;
      All.CPU_HydImbalance += sumimbalance / NTask;
    }
}


/*! This function is the 'core' of the SPH force computation. A target
 *  particle is specified which may either be local, or reside in the
 *  communication buffer.
 */
void hydro_evaluate(int target, int mode)
{
  int startnode, numngb;
  int j, k, n, timestep;
  FLOAT *pos, *vel;
  FLOAT mass, h_i, dhsmlDensityFactor, rho, pressure, f1, f2;
  double acc[3], dtEntropy, maxMu;
  double dx, dy, dz, dvx, dvy, dvz;
  double h_i2, hinv, hinv4;
  double p_over_rho2_i, p_over_rho2_j, soundspeed_i, soundspeed_j;
  double hfc, dwk_i, vdotr, vdotr2, visc, mu_ij, c_ij, rho_ij, h_ij;
  double h_j, dwk_j;
  double r, r2, u;
  double hfc_visc;
  double dmin1, dmin2;
  int imax1, imax2;

#ifdef CONDUCTION
  double condEnergyChange;
  double rEnergy_i, rEnergy_j, rEnergy_i_ns, rEnergy_j_ns, Entropy;
  double rKappa_i, rKappa_j;	/* Thermal Conduction Coefficients */
  double rEnergyFlow;
  double dtcond, smoothentr;

#ifdef CONDUCTION_SATURATION
  double gradentr[3];
  double electron_free_path, temp_scale_length;
#endif
#endif

#ifdef REDUCEVISC
  FLOAT alpha;
  double BulkVisc_ij;
#endif

#ifndef NOVISCOSITYLIMITER
  double dt;
#endif

#ifdef MAGNETIC
  FLOAT *bpred;
  double dtB[3];
  double magfac, magfac_i, magfac_j, magfac_i_base;

#ifdef MAGFORCE
  double mm_i[3][3], mm_j[3][3];
  double b2;
  int l;

#ifdef ARTBPRES
  double wk_i, wk_j, hinv3, aop = 8 / M_PI;
  double w1_i, w1_j, R_abp_i, R_abp_j;
  double mm_abp_i[3][3], mm_abp_j[3][3];
#endif
#endif

#ifdef TRACEDIVB
  double divB;
#endif

#endif

#ifdef SFR_DECOUPLING
  FLOAT densityold, entropy;
#endif


#ifdef PARTICLE_DEBUG
  int4byte ID;
#endif


  if(mode == 0)
    {
      pos = P[target].Pos;
      vel = SphP[target].VelPred;
      h_i = PPP[target].Hsml;
      mass = P[target].Mass;
#ifndef NOGRADHSML
      dhsmlDensityFactor = SphP[target].DhsmlDensityFactor;
#endif
      rho = SphP[target].Density;
      pressure = SphP[target].Pressure;
      timestep = P[target].Ti_endstep - P[target].Ti_begstep;
      soundspeed_i = sqrt(GAMMA * pressure / rho);
      f1 = fabs(SphP[target].DivVel) /
	(fabs(SphP[target].DivVel) + SphP[target].CurlVel +
	 0.0001 * soundspeed_i / PPP[target].Hsml / fac_mu);
#ifdef MAGNETIC
      bpred = SphP[target].BPred;
#endif
#ifdef REDUCEVISC
      alpha = SphP[target].alpha;
#endif
#ifdef CONDUCTION
      smoothentr = SphP[target].SmoothedEntr;
      Entropy = SphP[target].Entropy;
#ifdef CONDUCTION_SATURATION
      gradentr[0] = SphP[target].GradEntr[0];
      gradentr[1] = SphP[target].GradEntr[1];
      gradentr[2] = SphP[target].GradEntr[2];
#endif
#endif


#ifdef SFR_DECOUPLING
      densityold = SphP[target].DensityOld;
      entropy = SphP[target].Entropy;
#endif

#ifdef PARTICLE_DEBUG
      ID = SphP[target].ID;
#endif

    }
  else
    {
      pos = HydroDataGet[target].Pos;
      vel = HydroDataGet[target].Vel;
      h_i = HydroDataGet[target].Hsml;
      mass = HydroDataGet[target].Mass;
#ifndef NOGRADHSML
      dhsmlDensityFactor = HydroDataGet[target].DhsmlDensityFactor;
#endif
      rho = HydroDataGet[target].Density;
      pressure = HydroDataGet[target].Pressure;
      timestep = HydroDataGet[target].Timestep;
      soundspeed_i = sqrt(GAMMA * pressure / rho);
      f1 = HydroDataGet[target].F1;
#ifdef MAGNETIC
      bpred = HydroDataGet[target].BPred;
#endif
#ifdef REDUCEVISC
      alpha = HydroDataGet[target].alpha;
#endif
#ifdef CONDUCTION
      smoothentr = HydroDataGet[target].SmoothedEntr;
      Entropy = HydroDataGet[target].Entropy;
#ifdef CONDUCTION_SATURATION
      gradentr[0] = HydroDataGet[target].GradEntr[0];
      gradentr[1] = HydroDataGet[target].GradEntr[1];
      gradentr[2] = HydroDataGet[target].GradEntr[2];
#endif
#endif

#ifdef SFR_DECOUPLING
      densityold = HydroDataGet[target].DensityOld;
      entropy = HydroDataGet[target].Entropy;
#endif

#ifdef PARTICLE_DEBUG
      ID = HydroDataGet[target].ID;
#endif
    }


  /* initialize variables before SPH loop is started */

  acc[0] = acc[1] = acc[2] = dtEntropy = 0;
  maxMu = 0;
#ifdef CONDUCTION
  condEnergyChange = 0;
  dtcond = timestep * All.Timebase_interval;
  if(timestep == 0)
    dtcond = 1;
#endif
#ifdef MAGNETIC
  for(k = 0; k < 3; k++)
    dtB[k] = 0;
#ifdef TRACEDIVB
  divB = 0;
#endif
#ifdef MAGFORCE
  magfac_i_base = 1 / (rho * rho);
#ifndef BRIOWU
  magfac_i_base /= (4 * M_PI);
#endif
#ifdef CORRECTBFRC
  magfac_i_base *= dhsmlDensityFactor;
#endif
  for(k = 0, b2 = 0; k < 3; k++)
    {
      b2 += bpred[k] * bpred[k];
      for(l = 0; l < 3; l++)
	mm_i[k][l] = bpred[k] * bpred[l];
    }

  for(k = 0; k < 3; k++)
    mm_i[k][k] -= 0.5 * b2;
#endif
#endif /* end of MAGNETIC */
  p_over_rho2_i = pressure / (rho * rho);
#ifndef NOGRADHSML
  p_over_rho2_i *= dhsmlDensityFactor;
#endif
  h_i2 = h_i * h_i;
#ifdef CONDUCTION
  /* this gives the thermal energy per unit mass for particle i */
  rEnergy_i = smoothentr * pow(rho * a3inv, GAMMA_MINUS1) / GAMMA_MINUS1;
  rEnergy_i_ns = Entropy * pow(rho * a3inv, GAMMA_MINUS1) / GAMMA_MINUS1;
#ifdef CONDUCTION_CONSTANT
  rKappa_i = All.ConductionCoeff;
#else
  rKappa_i = All.ConductionCoeff * pow(rEnergy_i_ns, 2.5);
#ifdef CONDUCTION_SATURATION
  electron_free_path = All.ElectronFreePathFactor * rEnergy_i * rEnergy_i / (rho * a3inv);
  temp_scale_length =
    atime * fabs(smoothentr) / sqrt(gradentr[0] * gradentr[0] +
				    gradentr[1] * gradentr[1] + gradentr[2] * gradentr[2]);
  rKappa_i /= (1 + 4.2 * electron_free_path / temp_scale_length);
#endif
#endif
#ifdef SFR
  if(rho * a3inv >= All.PhysDensThresh)
    rKappa_i = 0;
#endif
#endif
  /* Now start the actual SPH computation for this particle */
  startnode = All.MaxPart;
  do
    {
#ifdef SFR_DECOUPLING
      numngb = ngb_treefind_pairs(&pos[0], h_i, &startnode, densityold, entropy, &vel[0]);
#else
      numngb = ngb_treefind_pairs(&pos[0], h_i, &startnode);
#endif
      for(n = 0; n < numngb; n++)
	{
	  j = Ngblist[n];

#ifdef BLACK_HOLES
	      if(P[j].Mass == 0)
		continue;
#endif

#ifdef WINDS
	  if(SphP[j].DelayTime > 0)	/* ignore the wind particles */
	    continue;
#endif
	  dx = pos[0] - P[j].Pos[0];
	  dy = pos[1] - P[j].Pos[1];
	  dz = pos[2] - P[j].Pos[2];
#ifdef PERIODIC			/*  now find the closest image in the given box size  */
#ifdef LONGBOX
	  if(dx > boxHalf * LONGBOX)
	    dx -= boxSize * LONGBOX;
	  if(dx < -boxHalf * LONGBOX)
	    dx += boxSize * LONGBOX;
#else
	  if(dx > boxHalf)
	    dx -= boxSize;
	  if(dx < -boxHalf)
	    dx += boxSize;
#endif
	  if(dy > boxHalf)
	    dy -= boxSize;
	  if(dy < -boxHalf)
	    dy += boxSize;
	  if(dz > boxHalf)
	    dz -= boxSize;
	  if(dz < -boxHalf)
	    dz += boxSize;
#endif
	  r2 = dx * dx + dy * dy + dz * dz;
	  h_j = PPP[j].Hsml;
	  if(r2 < h_i2 || r2 < h_j * h_j)
	    {
	      r = sqrt(r2);
	      if(r > 0)
		{
		  p_over_rho2_j = SphP[j].Pressure / (SphP[j].Density * SphP[j].Density);
		  soundspeed_j = sqrt(GAMMA * p_over_rho2_j * SphP[j].Density);
		  dvx = vel[0] - SphP[j].VelPred[0];
		  dvy = vel[1] - SphP[j].VelPred[1];
		  dvz = vel[2] - SphP[j].VelPred[2];
		  vdotr = dx * dvx + dy * dvy + dz * dvz;

		  if(All.ComovingIntegrationOn)
		    vdotr2 = vdotr + hubble_a2 * r2;
		  else
		    vdotr2 = vdotr;
		  if(r2 < h_i2)
		    {
		      hinv = 1.0 / h_i;
		      hinv4 = hinv * hinv * hinv * hinv;
		      u = r * hinv;
		      if(u < 0.5)
			dwk_i = hinv4 * u * (45.836623610466 * u - 30.557749073644);
		      else
			dwk_i = hinv4 * (-15.278874536822) * (1.0 - u) * (1.0 - u);
#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
		      hinv3 = hinv * hinv * hinv;
		      if(u <= 0.5)
			wk_i = aop * hinv3 * (1 - 6 * u * u * (1 - u));
		      else
			wk_i = aop * hinv3 * 2 * (1 - u) * (1 - u) * (1 - u);
		      if(u1 <= 0.5)
			w1_i = aop * hinv3 * (1 - 6 * u1 * u1 * (1 - u1));
		      else
			w1_i = aop * hinv3 * 2 * (1 - u1) * (1 - u1) * (1 - u1);
#endif
		    }
		  else
		    {
		      dwk_i = 0;
#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
		      wk_i = 0;
		      w1_i = 1;
#endif
		    }

		  if(r2 < h_j * h_j)
		    {
		      hinv = 1.0 / h_j;
		      hinv4 = hinv * hinv * hinv * hinv;
		      u = r * hinv;
		      if(u < 0.5)
			dwk_j = hinv4 * u * (45.836623610466 * u - 30.557749073644);
		      else
			dwk_j = hinv4 * (-15.278874536822) * (1.0 - u) * (1.0 - u);
#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
		      hinv3 = hinv * hinv * hinv;
		      if(u <= 0.5)
			wk_j = aop * hinv3 * (1 - 6 * u * u * (1 - u));
		      else
			wk_j = aop * hinv3 * 2 * (1 - u) * (1 - u) * (1 - u);
		      if(u1 <= 0.5)
			w1_j = aop * hinv3 * (1 - 6 * u1 * u1 * (1 - u1));
		      else
			w1_j = aop * hinv3 * 2 * (1 - u1) * (1 - u1) * (1 - u1);
#endif
		    }
		  else
		    {
		      dwk_j = 0;
#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
		      wk_j = 0;
		      w1_j = 1.;
#endif
		    }

#if defined(MAGNETIC) && defined(MAGFORCE) && defined(ARTBPRES)
		  R_abp_i = 0.3 * pow(wk_i / w1_i, 4);
		  R_abp_j = 0.3 * pow(wk_j / w1_j, 4);
#endif
		  if(vdotr2 < 0)	/* ... artificial viscosity */
		    {
		      c_ij = 0.5 * (soundspeed_i + soundspeed_j);
		      h_ij = 0.5 * (h_i + h_j);
		      mu_ij = fac_mu * h_ij * vdotr2 / (r2 + 0.01 * h_ij * h_ij);

		      if(maxMu < (-mu_ij))
			maxMu = (-mu_ij);

		      rho_ij = 0.5 * (rho + SphP[j].Density);
		      f2 =
			fabs(SphP[j].DivVel) / (fabs(SphP[j].DivVel) + SphP[j].CurlVel +
						0.0001 * soundspeed_j / fac_mu / PPP[j].Hsml);
#ifdef REDUCEVISC
		      BulkVisc_ij = 0.5 * (alpha + SphP[j].alpha);
		      visc =
			(-BulkVisc_ij * mu_ij * c_ij + 2 * BulkVisc_ij * mu_ij * mu_ij) / rho_ij * (f1 +
												    f2) * 0.5;
#else
		      visc =
			(-All.ArtBulkViscConst * mu_ij * c_ij +
			 2 * All.ArtBulkViscConst * mu_ij * mu_ij) / rho_ij * (f1 + f2) * 0.5;
#endif
		      /* .... end artificial viscosity evaluation */
		      /* now make sure that viscous acceleration is not too large */
#ifndef NOVISCOSITYLIMITER
		      dt = 2 * IMAX(timestep, (P[j].Ti_endstep - P[j].Ti_begstep)) * All.Timebase_interval;
		      if(dt > 0 && (dwk_i + dwk_j) < 0)
			{
#ifdef BLACK_HOLES
			      if((mass + P[j].Mass) > 0)
#endif
			  visc = DMIN(visc, 0.3 * fac_vsic_fix * vdotr2 /
				      (0.5 * (mass + P[j].Mass) * (dwk_i + dwk_j) * r * dt));
			}
#endif
		    }
		  else
		    visc = 0;
#ifndef NOGRADHSML
		  p_over_rho2_j *= SphP[j].DhsmlDensityFactor;
#endif
		  hfc_visc = 0.5 * P[j].Mass * visc * (dwk_i + dwk_j) / r;
		  /* Formulation derived from the Lagrangian */
		  hfc = hfc_visc + P[j].Mass * (p_over_rho2_i * dwk_i + p_over_rho2_j * dwk_j) / r;
#ifndef NOACCEL
		  acc[0] -= hfc * dx;
		  acc[1] -= hfc * dy;
		  acc[2] -= hfc * dz;
#endif
		  dtEntropy += 0.5 * hfc_visc * vdotr2;
#ifdef CONDUCTION
		  rEnergy_j =
		    SphP[j].SmoothedEntr * pow(SphP[j].Density * a3inv, GAMMA_MINUS1) / GAMMA_MINUS1;
		  rEnergy_j_ns = SphP[j].Entropy * pow(SphP[j].Density * a3inv, GAMMA_MINUS1) / GAMMA_MINUS1;
#ifdef CONDUCTION_CONSTANT
		  rKappa_j = All.ConductionCoeff;
#else
		  rKappa_j = All.ConductionCoeff * pow(rEnergy_j_ns, 2.5);
#ifdef CONDUCTION_SATURATION
		  electron_free_path =
		    All.ElectronFreePathFactor * rEnergy_j * rEnergy_j / (SphP[j].Density * a3inv);
		  temp_scale_length =
		    atime * fabs(smoothentr) / sqrt(SphP[j].GradEntr[0] * SphP[j].GradEntr[0] +
						    SphP[j].GradEntr[1] * SphP[j].GradEntr[1] +
						    SphP[j].GradEntr[2] * SphP[j].GradEntr[2]);
		  rKappa_j /= (1 + 4.2 * electron_free_path / temp_scale_length);
#endif
#endif
#ifdef SFR
		  if(SphP[j].Density * a3inv >= All.PhysDensThresh)
		    rKappa_j = 0;
#endif
		  if((rKappa_i + rKappa_j) > 0)
		    rEnergyFlow =
		      (mass * P[j].Mass) /
		      (rho * SphP[j].Density) *
		      4 * (rKappa_i * rKappa_j) / (rKappa_i + rKappa_j) * (rEnergy_j -
									   rEnergy_i) * 0.5 * (dwk_i +
											       dwk_j) / r *
		      dtcond;
		  else
		    rEnergyFlow = 0;
		  if(timestep > 0)
		    SphP[j].CondEnergyChange += 0.5 * rEnergyFlow;
		  condEnergyChange += -0.5 * rEnergyFlow;
#endif /* end of CONDUCTION */
#ifdef MAGNETIC
		  magfac = P[j].Mass / rho * dwk_i / r;
		  if(All.ComovingIntegrationOn)
		    magfac *= 1. / (hubble_a * All.Time * All.Time);
		  /* last factor takes care of all cosmological prefactor */
#ifdef CORRECTDB
		  magfac *= dhsmlDensityFactor;
#endif
		  dtB[0] +=
		    magfac * ((bpred[0] * dvy - bpred[1] * dvx) * dy +
			      (bpred[0] * dvz - bpred[2] * dvx) * dz);
		  dtB[1] +=
		    magfac * ((bpred[1] * dvz - bpred[2] * dvy) * dz +
			      (bpred[1] * dvx - bpred[0] * dvy) * dx);
		  dtB[2] +=
		    magfac * ((bpred[2] * dvx - bpred[0] * dvz) * dx +
			      (bpred[2] * dvy - bpred[1] * dvz) * dy);
#ifdef TRACEDIVB
		  divB += magfac *
		    ((SphP[j].BPred[0] - bpred[0]) * dx +
		     (SphP[j].BPred[1] - bpred[1]) * dy + (SphP[j].BPred[2] - bpred[2]) * dz);
#endif
#ifdef MAGFORCE
		  magfac_j = 1 / (SphP[j].Density * SphP[j].Density);
#ifndef BRIOWU
		  magfac_j /= (4 * M_PI);
#endif
#ifdef CORRECTBFRC
		  magfac_j *= dwk_j * SphP[j].DhsmlDensityFactor;
		  magfac_i = dwk_i * magfac_i_base;
#else
		  magfac_i = magfac_i_base;
#endif
		  for(k = 0, b2 = 0; k < 3; k++)
		    {
		      b2 += SphP[j].BPred[k] * SphP[j].BPred[k];
		      for(l = 0; l < 3; l++)
			mm_j[k][l] = SphP[j].BPred[k] * SphP[j].BPred[l];
		    }

		  for(k = 0; k < 3; k++)
		    mm_j[k][k] -= 0.5 * b2;
#ifdef CORRECTBFRC
		  magfac = P[j].Mass / r;
#else
		  magfac = P[j].Mass * 0.5 * (dwk_i + dwk_j) / r;
#endif
		  if(All.ComovingIntegrationOn)
		    magfac *= pow(All.Time, 3 * GAMMA);
		  /* last factor takes care of all cosmological prefactor */
#ifndef BRIOWU
		  magfac /= All.UnitVelocity_in_cm_per_s / All.UnitTime_in_s;
		  /* take care of B unit conversion into GADGET units ! */
#endif
		  for(k = 0; k < 3; k++)
		    acc[k] +=
		      magfac * ((mm_i[k][0] * magfac_i + mm_j[k][0] * magfac_j) * dx +
				(mm_i[k][1] * magfac_i + mm_j[k][1] * magfac_j) * dy +
				(mm_i[k][2] * magfac_i + mm_j[k][2] * magfac_j) * dz);
#ifdef ARTBPRES
		  for(k = 0; k < 3; k++)
		    for(l = 0; l < 3; l++)
		      mm_abp_j[k][l] = SphP[j].BPred[k] * SphP[j].BPred[l] * R_abp_j / 2.0;
		  for(k = 0; k < 3; k++)
		    for(l = 0; l < 3; l++)
		      mm_abp_i[k][l] = bpred[k] * bpred[l] * R_abp_i / 2.0;
		  for(k = 0; k < 3; k++)
		    acc[k] -= magfac * ((mm_abp_i[k][0] * magfac_i +
					 mm_abp_j[k][0] * magfac_j) * dx +
					(mm_abp_i[k][1] * magfac_i +
					 mm_abp_j[k][1] * magfac_j) * dy +
					(mm_abp_i[k][2] * magfac_i + mm_abp_j[k][2] * magfac_j) * dz);
#endif
#ifdef DIVBFORCE
		  for(k = 0; k < 3; k++)
		    acc[k] -=
		      magfac * (((bpred[k] * bpred[0]) * magfac_i
				 + (bpred[k] * SphP[j].BPred[0]) * magfac_j) * dx
				+ ((bpred[k] * bpred[1]) * magfac_i
				   + (bpred[k] * SphP[j].BPred[1]) * magfac_j) * dy
				+ ((bpred[k] * bpred[2]) * magfac_i +
				   (bpred[k] * SphP[j].BPred[2]) * magfac_j) * dz);
#endif
#endif
#endif /* end of MAGNETIC */
		}
	    }
	}
    }
  while(startnode >= 0);
  /* Now collect the result at the right place */
  if(mode == 0)
    {
      for(k = 0; k < 3; k++)
	SphP[target].HydroAccel[k] = acc[k];
      SphP[target].DtEntropy = dtEntropy;
      SphP[target].MaxMu = maxMu;
#ifdef CONDUCTION
      SphP[target].CondEnergyChange += condEnergyChange;
#ifdef OUTPUTCOOLRATE
      SphP[target].CondRate = 2 * atime * condEnergyChange / (dtcond * mass);
#endif
#endif
#ifdef MAGNETIC
      for(k = 0; k < 3; k++)
	SphP[target].DtB[k] = dtB[k];
#ifdef TRACEDIVB
      SphP[target].divB = divB;
#endif
#endif
    }
  else
    {
      for(k = 0; k < 3; k++)
	HydroDataResult[target].Acc[k] = acc[k];
      HydroDataResult[target].DtEntropy = dtEntropy;
      HydroDataResult[target].MaxMu = maxMu;
#ifdef CONDUCTION
      HydroDataResult[target].CondEnergyChange = condEnergyChange;
#ifdef OUTPUTCOOLRATE
      HydroDataResult[target].CondRate = 2 * atime * condEnergyChange / (dtcond * mass);
#endif
#endif
#ifdef MAGNETIC
      for(k = 0; k < 3; k++)
	HydroDataResult[target].DtB[k] = dtB[k];
#ifdef TRACEDIVB
      HydroDataResult[target].divB = divB;
#endif
#endif
    }
}




/*! This is a comparison kernel for a sort routine, which is used to group
 *  particles that are going to be exported to the same CPU.
 */
int hydro_compare_key(const void *a, const void *b)
{
  if(((struct hydrodata_in *) a)->Task < (((struct hydrodata_in *) b)->Task))
    return -1;
  if(((struct hydrodata_in *) a)->Task > (((struct hydrodata_in *) b)->Task))
    return +1;
  return 0;
}
